# Harmonize

- [Overview](#overview)
- [What does it do?](#what-does-it-do)
- [Should I use it?](#should-i-use-it)
- [Quick Start](#quick-start)
- [Features](#features)
  - [Module's Main Settings](#modules-main-settings)
  - [Entity Settings](#entity-settings)
  - [Data Alteration](#data-alteration)
  - [Twig Functions & Filters](#twig-functions--filters)
  - [Entity Processing Rules](#entity-processing-rules)
  - [Styles](#styles)
  - [Caching](#caching)
  - [Visualizer](#visualizer)

## Overview
Harmonize streamlines preprocessing of entities in Drupal before template rendering.
It is a tool for developers and was originally made to facilitate collaboration between backend and frontend developers.

## What does it do?
Harmonize makes it much easier to access field values in Drupal templates. It does this by preprocessing entity content
and adding a new variable called `harmony` to your twig templates. This variable contains a simple, cleanly formatted
array of data to work with.

Drupal already exposes the values of your content to your templates, but it is sometimes difficult to access said data.
This is especially true when trying to access values in nested entities.

Take this example. Here we're fetching the first 3 articles entered in a `field_related_articles` field of a `node` template.
After fetching these articles, we access their thumbnails and want to print their path, alt and title values.

### Standard

```html
<div>
    {% for articleFieldValue in node.field_related_articles[0:3] %}
        <img
            src="{{ file_url(articleFieldValue.entity.field_thumbnail.entity.field_media_image.entity.fileuri) }}"
            alt="{{ articleFieldValue.entity.field_thumbnail.entity.field_media_image.alt }}"
            title="{{ articleFieldValue.entity.field_thumbnail.entity.field_media_image.title }}"
        />
    {% endfor %}
</div>
```

### With Harmonize (Out of the Box)

```html
<div>
    {% for article in harmony.field_related_articles[0:3] %}
        <img
            src="{{ article|get('field_thumbnail').field_media_image.attributes.src }}"
            alt="{{ article|get('field_thumbnail').field_media_image.attributes.alt }}"
            title="{{ article|get('field_thumbnail').field_media_image.attributes.title }}"
        />
    {% endfor %}
</div>
```

If this already makes you a little happier, then good! However, there are many more features that can further
streamline preprocessing on your Drupal website. We will cover this in additional examples below.

## Should I use it?

The best use case for the Harmonize module is when your frontend needs to branch out of the traditional Drupal way of
handling things. Standard Drupal websites can make good use of the current templating system and render arrays, taking
full advantage of Display Modes and much more. However, if your website needs a more flexible frontend solution you simply
need to get the data from Drupal entities, this module can help you achieve that.

To get a good idea of a great use case: This module was initially conceived to facilitate development in a project where
the frontend developers were working in a custom styleguide with its own templates. This module was also used to bring
together [Fractal](https://fractal.build/) and Drupal sites.

Otherwise, this module out-of-the-box can still facilitate preprocessing on your Drupal website.

## Quick Start

To see how this module works out of the box, you can simply enable it and set up simple content configurations.

Once the module is enabled, edit any entity type. In the example below, I am editing a **Content Type** called **Page**.

![Content Type Settings](./doc/screenshots/Screenshot_1.png)

Check the first checkbox shown in the screenshot to enable the module to preprocess data and assign it to your templates.

Optionally, you can check the second checkbox and then visit your content type's display settings to confirm which fields are preprocessed.

By default, all fields are processed.

![Content Type Display Settings](./doc/screenshots/Screenshot_2.png)

Finally, you can now inspect the contents of the `harmony` variable either inside a `node.html.twig` or in a `hook_preprocess_node` function.

In this example, we'll opt for the latter to avoid any memory errors caused by Twig's default dumper. (If you're on Drupal 9.5+, you can safely use `dump()` in Templates).

```php
/**
 * Implements hook_preprocess_HOOK() for node.html.twig.
 */
function dreamland_theme_preprocess_node(&$variables) {
  dump($variables['harmony']);
}
```

Or in Twig (Safe to do in Drupal 9.5.x+)

```html
{{ dump(harmony) }}
```

The data should be shown to you when visiting your page. Here is an example of the data in this example.

![Data Example 1](./doc/screenshots/Screenshot_3.png)

And you're done! Now this is the out-of-the-box configuration, but there are ways to optimize the data even further.

For instance, you may notice that in the data example shown here, the entity reference fields return an unfamiliar class.

These classes are defined in the module and are called **Harmonizers**. But more on this further down the line.

As you go through the guide, the data examples will become more and more compact and easy to navigate. This will be done through
the use of many of the module's features.

## Upgrading from V2 to V3

So you're working on a website that has V2 installed, and you want to upgrade to V3...

The first thing to note is that there are many differences in V3. One of the main things is that nested preprocessing of
entities in entity reference fields no longer happens automatically. This is because V2 had a lot of risks with this
behaviour. Infinity loops were much more prevalent.

With this major change, you will have to refactor most of your existing code pertaining to harmonize.

When `entity_reference` fields are processed, a **Harmonizer** is returned instead of the expected data of the referenced
entities. To reach the nested data, you simply have to call `->harmonize()` on the Harmonizer object. Keep this in mind
when refactoring your code.

### Upgrade Steps

Follow these steps to assure the least amount of headache when upgrading.

* Make sure you site configurations in your codebase are up-to-date: `drush cex -y`.
* Update the module via composer: `composer update drupal/harmonize`.
* Run database updates: `drush updb -y`.
* Export configurations. This is important as the database updates migrate old configurations to new ones.
* Now you should be good to go.

### Changelog

Below are other things to consider when upgrading to V3.

* As mentioned above, nested processing of referenced entities is no longer automatic. Adjust code accordingly.
* Harmonizer Signatures have been fine-tuned. When targeting entities in a specific style, you must adjust your harmony file names and Refiner targets.
  * `entity.node.article--card` becomes `entity.node.article.card`.
* Data configurations have been deprecated. They have been replaced with [Entity Processing Rules](#entity-processing-rules).
* [Styles](#styles) can now be created in the admin interface.
* A [Visualizer](#visualizer) has been added to allow entity processing tests.
* New [Twig](#twig-functions--filters) functionality has been added. This may help with refactoring.
* Caching has been optimized and is no longer experimental. Enable it.
* [Entity Settings](#entity-settings) have been enabled along with view mode integration.
* [Main Module Settings](#modules-main-settings) have been adjusted.
  * Moved Automatic Preprocessing setting to a textarea. You now specify which entity types are automatically processed in entity reference fields.
  * Removed recursive depth settings.
  * Added new setting for data alteration object load. See [this section](#data-alteration) for more details.

## Features

### Module's Main Settings

The main settings of the module can be configured at `/admin/structure/harmonize/settings`.

![Main Settings](./doc/screenshots/Screenshot_4.png)
![Main Settings 2](./doc/screenshots/Screenshot_5.png)

Below is a description of each setting.

#### Harmony File Discovery

This enables [Harmony File Discovery](#harmony-files). Refer to the [section](#harmony-files) specific to this below.

#### Loading Algorithm

This configures the loading algorithm. Don't mind this for now as this will be covered in the
[Data Alteration](#data-alteration) section.

#### Automatically Processed Entities

This allows you to configure which entities are automatically processed. If you remember the [Quick Start Example](#quick-start)
above, you saw that entity reference fields values return a **Harmonizer**. If the entity type being processed is in this
setting, it will return the processed data directly instead.

Here is an example of the data from the Quick Start above if `node.article` is added in this setting.

![Data Example 2](./doc/screenshots/Screenshot_6.png)

Instead of returning a **Harmonizer**, we now see the data of the referenced entities in the entity reference fields.

The module normally returns **Harmonizers** to avoid surprise infinity loops in entity reference fields. Say an article in my example had an entity
reference field of its own pointing towards a page, there would be risk of infinity loops if entities were automatically processed.

With that in mind, be wary of the entities you add to this setting.

You can always accomplish this manually by doing something similar to this:

```php
/**
 * Implements hook_preprocess_HOOK() for node.html.twig.
 */
function dreamland_theme_preprocess_node(&$variables) {
  // Manually process nested entities.
  $variables['harmony']['field_related_articles'] = array_map(function ($articleHarmonizer) {
    return $articleHarmonizer->harmonize();
  }, $variables['harmony']['field_related_articles']);
}
```

All **Harmonizers** can be used to access nested data with the `->harmonize()` function.

#### Allowed Admin Routes

Harmonize normally does not execute in admin routes as it is defined to act before the templating layer of the frontend. With that
being said, there are cases where you do want it to execute in an admin route. Layout builder paths are a good example of this since
to preview blocks while editing content, Drupal needs to render the entities for you to see.

Note that the layout builder routes in question are currently already hardcoded, so you don't need to add them yourself.
Chances are this setting won't need to be touched for a while.

### Entity Settings

As seen in the [Quick Start](#quick-start), you have access to entity specific settings. These will usually be seen when
editing an entity type.

![Content Type Settings](./doc/screenshots/Screenshot_1.png)

#### Populate Theme Variables with Harmonized Data

This quite simply adds the `harmony` key to the `$variables` array of your entity's template.

#### Enable display filtering

Grants access to filtering which fields are preprocessed per display type. See the screenshot below.

![Content Type Display Settings](./doc/screenshots/Screenshot_2.png)

New fields will be included by default. You simply uncheck the checkbox on any field you want to omit. Disabled fields
with the checkbox checked will still be preprocessed.

### Data Alteration

Now we're getting to the good stuff!

Everything we've seen so far is what can be done without any additional code. Harmonize offers way to fine-tune the data
before it is sent to templates.

Now the easiest way of course would be through your traditional Drupal Hooks.

```php
/**
 * Implements hook_preprocess_HOOK() for node.html.twig.
 */
function dreamland_theme_preprocess_node(&$variables) {
  // Get the data from Harmonize.
  $harmony = &$variables['harmony'];

  // Do data alteration stuff...
  $harmony['related_article_count'] = count($harmony['field_related_articles']);
}
```

This is completely fine to do, but Harmonize offers other ways to do this without writing hooks. The two main ways to do
this are:

* Harmony Files
* Refiners

Henceforth, these will be referred to as **Alteration Objects**.

#### Harmony Files

**Harmony Files** are the main way to alter data before it is sent to templates. You need the [**Harmony File Discovery**](#harmony-file-discovery) setting
to be enabled for this feature to be available to you.

With the feature enabled, Harmonize will look for files in your active theme. The file crawling is similar to how Drupal's templating
system works.

**e.g.** For a `node` of type `page`, you would create a `node--page.html.twig` in the `templates` directory of your theme.

**cont.** For the same `node` of type `page`, you would create a `entity.node.page.harmony.php` file in the `harmony` directory of your theme.

The contents of the `entity.node.page.harmony.php` file would be as follows:

```php
<?php

use Drupal\harmonize\Service\Harmonics;

/**
 * @file
 * Override & customize data harmonizations for a given Drupal object or entity.
 *
 * @harmonizerType entity
 * @entityType node
 * @bundle page
 *
 * This is a custom file to handle the reformatting of data before it reaches
 * Drupal templates.
 *
 * Available variables:
 * @var array $consignment
 *    The data consigned by core Harmonize behaviour or Refiners.
 * @var array $harmony
 *    The data that will be returned to the Twig template. If this is empty, $consignment will be returned.
 * @var Harmonics $harmonics
 *    The Harmonics service which contains helper functions.
 *
 * Re-arrange and assign data to the $harmony variable here, and that's all you
 * need. The rest will be handled in the Twig Templates.
 */

// The same alterations we did earlier, but in a harmony file!
$consignment['related_article_count'] = count($consignment['field_related_articles']);
```

The `$consignment` variable contains all the data so far. The `$harmony` variable is the data that will be returned to
templates. If `$harmony` is empty by the end of the file, `$consignment` will be returned. So you have two options in
Harmony Files:

* Modify `$consignment` and never set `$harmony`. This is good for small touch-ups of the data.
* Set `$harmony` variable with new data. This is good if you have to **rearrange** data or make bigger changes.

In our example above, we are simply adding new data, so we modify `$consignment`.

And voila! In your `node--page.html.twig` template, you should now have access to `harmony.related_article_count`.

Note that Harmony Files run before your traditional hooks, so you can always do a combination of both.

If the Harmony File above exists, then:

```php
/**
 * Implements hook_preprocess_HOOK() for node.html.twig.
 */
function dreamland_theme_preprocess_node(&$variables) {
  // Since the harmony file declared above exists, this should evaluate to true.
  if (isset($harmony['related_article_count'])) {
    // Yay! It exists!
  }
}
```

As you may have guessed, the file crawling works similar to template suggestions. So `entity.node.harmony.php` would work too,
except it would apply to all nodes instead of just `page` typed nodes. You can also specify a view mode, like `entity.node.page.default`
or `entity.node.page.teaser`.

#### Refiners

Refiners are alteration objects that work very similarly to harmony files. They are quite simply your OOP solution to them, allowing for Dependency Injection
and more.

Refiners use the [Plugin API](https://www.drupal.org/docs/drupal-apis/plugin-api).

If you haven't read the [Harmony Files](#harmony-files) section, I suggest you do so now to get a good comparison!

First, you would create a plugin in a module. For the sake of this example, I am creating a module called `dreamland_refiners`.

Then, I can create a plugin at `dreamland_refiners/src/Plugin/harmonize/Refiner/PageNodeRefiner.php`.

The contents would be the following:

```php
<?php

namespace Drupal\dreamland_refiners\Plugin\harmonize\Refiner;

use Drupal\harmonize\PluginManager\Refinery\RefinerBase;

/**
 * Provide plugin to alter harmonizer data for an entity.
 *
 * @noinspection
 *   AnnotationMissingUseInspection
 *
 * @Refiner(
 *   id = "dreamland_refiners.entity.node.page.refiner",
 *   target = "entity.node.page"
 * )
 *
 * @package Drupal\harmonize\Plugin\Harmonizer\Refiner
 */
final class PageNodeRefiner extends RefinerBase {

  /**
   * Add personal tweaks to harmonized data in this function.
   *
   * {@inheritdoc}
   */
  public function refine(&$consignment, &$harmony, $object): void {
    // Similar alterations to what we did earlier, but in a refiner!
    $consignment['featured_article_count'] = count($consignment['field_featured_articles']);
  }

}
```

Refiners work quite exactly like Harmony Files. Anything found inside the `$harmony` variable at the end of the `refine()`
function's execution will be sent to templates. If `$harmony` isn't set, `$consignment` will be taken instead.

The refiner here knows what to target due to the annotation definition. You can see that the target is set to `entity.node.page`.

This works just like Harmony files. You could set it to `entity.node` to target all nodes, or `entity.node.page.default` for the
`default` display type.

#### Alteration Object Flow

So now we should have a good understanding of data alteration objects. With this, we can finally talk about what happens
when all of them exist.

Let's say we have everything from the examples above, so:

* A Refiner at `dreamland_refiners/src/Plugin/harmonize/Refiner/PageNodeRefiner.php` with the same content as above.
* A Harmony File at `dreamland_theme/harmony/entity.node.page.harmony.php` with the same content as above.

The final data would be the following:

![Data Example 3](./doc/screenshots/Screenshot_7.png)

But you're probably wondering what goes first? The answer to your question is right here.

`Harmonize Preprocesses Entity -> Refiners Alter Data -> Harmony Files Alter Data -> Data is sent to 'harmony' key in $variables -> Data is sent to twig templates`

This flow promotes the following best practices:

* Use Harmonize for easy base preprocessing of entities.
* Create Refiners for big data alterations that require advanced Drupal service injection, database queries and more.
* Create Harmony Files OR preprocess functions in your theme for final touch-ups of data before it reaches templates.
* Use `harmony` in your templates to access the preprocessed data.

#### Alteration Object Loading Algorithm Explained

Remember that [setting](#loading-algorithm) from above? Now we can finally talk about how it works.

Harmonize has two ways to load alteration objects. To showcase these two methodologies, let's say we have the following:

* A harmony file called `entity.node.harmony.php`.
* A harmony file called `entity.node.page.harmony.php`.
* We're loading a `node` of type `page`.

Refiners work in the same way, so any explanations below apply to them too.

##### FCFS (First Come First Serve)

**FCFS** will **only load the first alteration object found**. It attempts to load **the most relevant alteration object** for
a given template.

So in the case above, only `entity.node.page.harmony.php` will be loaded, because you're loading a `node` of type `page`.

Other harmony files will be ignored once one is loaded.

This loading philosophy is akin to core loading philosophies for templates for example.

This is also the **recommended** setting to use.

##### SCAN (Elevator)

**SCAN** will **load multiple alteration objects**. It loads **all relevant alteration objects** for a given template.

So in the case above, both `entity.node.page.harmony.php` and `entity.node.harmony.php` will be loaded. It will process
them in order of how specific your alteration object is, from least to most specific. Here, `entity.node.harmony.php`
will be loaded first, and then `entity.node.page.harmony.php` will be loaded after.

### Twig Functions & Filters

Harmonize currently offers one twig function and one twig filter.

#### harmonize()

The `harmonize()` filter can be used on an entity in a template to get harmonized data.

For example in `node.html.twig`, we usually have access to the `node` variable, which contains the node being viewed.

If you don't have pre-process hooks and want to harmonize the data, you can do the following.

```html
{% set harmony = harmonize(node) %}
```

This is something that you can do if you don't want Harmonize to do things automatically. You can also use this function
to get harmonized data from entities in Entity Reference fields.

#### |get('FIELD_NAME')

The `|get('FIELD_NAME')` filter can be used from a **Harmonizer** to get a field value from a Harmonizer.

This is the same example as the one at the top of this guide. Here, we use it to get harmonized data on the referenced
articles.

```html
<div>
    {% for article in harmony.field_related_articles[0:3] %}
        <img
            src="{{ article|get('field_thumbnail').field_media_image.attributes.src }}"
            alt="{{ article|get('field_thumbnail').field_media_image.attributes.alt }}"
            title="{{ article|get('field_thumbnail').field_media_image.attributes.title }}"
        />
    {% endfor %}
</div>
```

### Styles

**More info soon!**

### Entity Processing Rules

Entity Processing Rules can be managed `admin/structure/harmonize/entity-rules`.

The settings here allow you to specify certain behaviors when Harmonize processes fields. You can:

* Exclude a field from all processing.
* Disable caching on a field.
* Automatically process referenced entities (Entity Reference Fields Only)
* Set a Harmonizer Style (Entity Reference Fields Only)

This page is great for setting global processing rules for entities. For example, if there is a field that you never
want Harmonize to process, this is where you would set it.

**e.g.** Excluding all Layout Builder fields from processing in nodes.

You can set up rules for Entities and Entity Bundles.

### Caching

With caching enabled, Harmonize will cache preprocessing results. This is **heavily recommended** to keep on simply for
performance reasons.

Harmonize results will share cache tags with the object it processes.

You can configure the cache at `admin/structure/harmonize/cache`.

### Visualizer

The Visualizer is a powerful tool that can be used to test entity preprocessing.

You can configure the visualizer at `admin/structure/harmonize/visualize`.

**More information soon.**
