<?php

namespace Drupal\harmonize\EventSubscriber;

use Drupal\Component\Plugin\Exception\PluginException;
use Drupal\harmonize\Constants\HarmonizeConfig;
use Drupal\harmonize\Constants\HarmonizerFlags;
use Drupal\harmonize\Event\HarmonizationEvent;
use Drupal\harmonize\Constants\HarmonizationEvents;
use Drupal\harmonize\Harmonizer\EntityHarmonizer\EntityHarmonizerInterface;
use Drupal\harmonize\PluginManager\Refinery\RefinerInterface;
use Drupal\harmonize\PluginManager\Refinery\Refinery;
use Drupal\harmonize\Service\Harmonize;
use Drupal\harmonize\Service\Helpers;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;

/**
 * Subscribes events to apply Plugin Alterations of harmonization data.
 *
 * @property Refinery $refinery
 *
 * @package Drupal\custom_events\EventSubscriber
 */
class RefineryEventSubscriber implements EventSubscriberInterface {

  /**
   * The Helpers service injected through DI.
   *
   * @var Helpers
   */
  private Helpers $helpers;

  /**
   * Use DI to inject the Refinery Plugin Manager.
   *
   * @var Refinery
   */
  private Refinery $refinery;

  /**
   * A flag that determines whether we found a refiner.
   */
  private bool $loadedRefiner = FALSE;

  /**
   * RefineryEventSubscriber constructor.
   *
   * @param Helpers $helpers
   *   The Helpers service injected through DI.
   * @param Refinery $refinery
   *   Refinery Plugin Manager injected through DI.
   */
  public function __construct(Helpers $helpers, Refinery $refinery) {
    $this->helpers = $helpers;
    $this->refinery = $refinery;
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents() : array {
    return [
      HarmonizationEvents::CORE => ['onHarmonize', 200],
    ];
  }

  /**
   * React to a Harmonizer event.
   *
   * @param HarmonizationEvent $event
   *   Event that occurred, containing the harmonizer that fired the event.
   *
   * @throws PluginException
   */
  public function onHarmonize(HarmonizationEvent $event) : void {
    // Get the signature parts.
    /* @see \Drupal\harmonize\Harmonizer\Harmonizer::getSignatureParts() */
    $signatureParts = $event->harmonizer->getSignatureParts(TRUE);

    // If 'FCFS' is set, we only want to get the first refiner we find.
    if ($this->isFCFS()) {
      $event->harmonizer->trace(t('[Refinery Event] <strong>FCFS</strong> Algorithm Detected...Only a single Refiner will be loaded'));
      // Okay, cool stuff now.
      // We reverse the signature parts array before proceeding.
      // This is because we are only going to be trying to get a single refiner.
      $signatureParts = array_reverse($signatureParts);
    }

    // With each part of the signature, search for Refiners matching the part.
    foreach ($signatureParts as $part) {
      // The signature part should match the target attribute of the Refiner.
      $refinerTarget = $part;

      // Get all plugin definitions with the specified target signature.
      $refiners = $this->refinery->getRefiners($refinerTarget);

      // If there are no refiners, we can stop here.
      if (empty($refiners)) {
        $event->harmonizer->trace(t("[Refinery Event] No refiners found for <strong><code>[@part]</code></strong>...Continuing", ['@part' => $refinerTarget]));
        continue;
      }

      // At this point we know a refiner was found.
      $this->loadedRefiner = TRUE;

      // Sort the refiners by weight.
      uasort($refiners, 'self::sortRefiners');

      $event->harmonizer->trace(t("[Refinery Event] <strong>Found @count Refiner(s)</strong> for <strong><code>[@part]</code></strong>...Continuing", [
        '@count' => count($refiners),
        '@part' => $refinerTarget,
      ]));

      // If 'FCFS', we only want one refiner.
      if ($this->isFCFS()) {
        $event->harmonizer->trace(t("[Refinery Event] Due to <strong>FCFS</strong>, only the <strong><code>[@refinerId]</code></strong> Refiner will be used...Proceeding", [
          '@refinerId' => reset($refiners)->getId(),
        ]));
        $refiners = [reset($refiners)];
      }

      // Perform data alterations for all refiners.
      foreach ($refiners as $refiner) {
        $consignment = $event->harmonizer->getData();
        if ($refiner !== NULL) {
          $harmony = [];
          $event->harmonizer->trace(t("[Refinery Event] Altering data with Refiner <strong><code>[@refinerId]</code></strong>", [
            '@refinerId' => $refiner->getId(),
          ]));
          $refiner->refine($consignment, $harmony, $event->harmonizer->getObject());
          // If the harmony object is empty, that means it wasn't reformatted.
          // We'll return the original consignment.
          if (!empty($harmony)) {
            $event->harmonizer->setData($harmony);
            $event->harmonizer->trace(t("[Refinery Event] <strong><code>\$harmony</code></strong> variable was set...<strong><code>\$consignment</code></strong> data has been overwritten"), TRUE);
          }
          else {
            $event->harmonizer->setData($consignment);
            $event->harmonizer->trace(t("[Refinery Event] Data modified and returned with alterations from Refiner"), TRUE);
          }
        }
      }

      // If 'FCFS', we break at this point.
      if ($this->isFCFS()) {
        break;
      }
    }

    // If no files were loaded, we can stop here.
    if (!$this->loadedRefiner) {
      $event->harmonizer->trace(t("[Refinery Event] No refiners were found. Data will be returned in its current state"), TRUE);
    }
  }

  /**
   * Helper function to tell us if we're in 'FCFS' mode.
   *
   * @return bool
   *   Returns TRUE if we're in FCFS mode, returns FALSE otherwise.
   */
  private function isFCFS() : bool {
    return Harmonize::$config[HarmonizeConfig::MAIN__LOADING_ALGORITHM] === 'fcfs';
  }

  /**
   * Sort refiners according to their weight.
   *
   * @param RefinerInterface $a
   *   First refiner plugin.
   * @param RefinerInterface $b
   *   Second refiner plugin.
   *
   * @return int
   *   Returns a value less than, equal to or greater than zero depending on if
   *   the first argument is considered to be respectively less than, equal to
   *   or greater than the second.
   */
  private function sortRefiners(RefinerInterface $a, RefinerInterface $b) : int {
    return $a->getWeight() - $b->getWeight();
  }

}
