<?php

namespace Drupal\harmonize\Form\EntityRules;

use Drupal;
use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\ReplaceCommand;
use Drupal\Core\Entity\ContentEntityType;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;
use Drupal\field\FieldConfigInterface;
use Drupal\harmonize\Constants\HarmonizeConfig;
use Drupal\harmonize\Service\Helpers;

/**
 * ModalForm class.
 */
class HarmonizeAddRulesForm extends ConfigFormBase {

  /**
   * Property to store the step in this multistep form.
   *
   * @var int
   */
  protected int $step = 1;

  /**
   * {@inheritdoc}
   */
  public function getFormId() : string {
    return 'harmonize_admin_add_rules_form';
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames() : array {
    return [
      HarmonizeConfig::ENTITY_RULES,
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) : array {
    /**
     * Services.
     * @var Helpers $helpers
     * @TODO - We don't want to DI because it has weird results with AJAX. Find a solution to this.
     */
    $helpers = Drupal::service('harmonize.helpers');
    $entityTypeManager = Drupal::entityTypeManager();

    // Some setup.
    $form['#prefix'] = '<div id="harmonize_add_rules_modal_form">';
    $form['#suffix'] = '</div>';
    $form['#attached']['library'][] = 'core/drupal.dialog.ajax';

    // The status messages that will contain any form errors.
    $form['status_messages'] = [
      '#type' => 'status_messages',
      '#weight' => -10,
    ];

    // Actions.
    $form['actions'] = array('#type' => 'actions');

    // Manage step 1.
    if ($this->step == 1) {
      // Now we want to get ALL the entity types installed on the site and list them.
      $entityTypeDefinitions = $entityTypeManager->getDefinitions();

      // Get only content entity types.
      $contentEntityTypeDefinitions = array_filter($entityTypeDefinitions, function($definition) {
        return $definition instanceof ContentEntityType;
      });

      // Create and populate our options array.
      $contentEntityTypeOptions = [];
      foreach ($contentEntityTypeDefinitions as $definition) {
        $id = $definition->id();
        $label = $definition->getLabel()->getUntranslatedString();
        $contentEntityTypeOptions[$id] = $label;
      }
      asort($contentEntityTypeOptions);

      $form['entity_type'] = [
        '#title' => t('Entity Type'),
        '#type' => 'select',
        // '#required' => TRUE,
        '#description' => 'Select the entity type this rule will be created for.',
        '#options' => $contentEntityTypeOptions,
        '#attributes' => [
          'class' => [
            'use-ajax',
          ],
        ],
        '#ajax' => [
          'callback' => '::addBundlesSelection',
          'wrapper' => 'js-ajax-bundle-wrapper',
          'disable-refocus' => TRUE,
        ],
      ];

      $form['bundle'] = [
        '#title' => t('Bundle'),
        '#type' => 'select',
        '#description' => 'Select the entity type this rule will be created for.',
        '#prefix' => '<div id="js-ajax-bundle-wrapper">',
        '#suffix' => '</div>',
        '#options' => [
          '' => $this->t('--'),
        ],
        "#validated" => TRUE,
      ];

      $form['actions']['continue'] = [
        '#type' => 'submit',
        '#name' => 'continue',
        '#value' => $this->t('Continue'),
        '#attributes' => [
          'class' => [
            'use-ajax',
          ],
        ],
        '#ajax' => [
          'callback' => '::ajax',
          'wrapper' => 'harmonize_add_rules_modal_form',
          'disable-refocus' => TRUE,
        ],
      ];
    }

    // Manage step 2.
    if ($this->step == 2) {
      // Get the selected entity type and bundle.
      $entityType = $form_state->getValue('entity_type');
      $bundle = $form_state->getValue('bundle');

      // Get all fields for this entity type.
      // @TODO - Can't DI cause AJAX messes with it. Find solution to that.
      if (empty($bundle)) {
        $fields = $helpers->getEntityTypeFieldDefinitions($entityType);
      } else {
        $fields = $helpers->getEntityTypeBundleFieldDefinitions($entityType, $bundle);
      }

      // We build a list of fields.
      /* @var FieldConfigInterface $field */
      foreach ($fields as $field) {
        $form['fields'][$field->id()] = [
          '#type' => 'container',
          '#attributes' => [
            'class' => ['container-inline'],
          ],
        ];
        $form['fields'][$field->id()]['label']['#markup'] = $field->label() . " <small>(<code>" . $field->id() . "</code>)</small> ";
        $form['fields'][$field->id()]['add'] = [
          '#type' => 'submit',
          '#name' => 'add_field_' . $field->id(),
          '#value' => $this->t('Add'),
          '#submit' => ['::addField'],
          '#field' => $field,
          '#entityType' => $entityType,
          '#bundle' => $bundle,
          '#attributes' => [
            'class' => [
              'use-ajax',
              'button',
              'button-action',
              'button--primary',
              'button--small',
            ],
          ],
          '#ajax' => [
            'callback' => '::ajax',
            'wrapper' => 'harmonize_add_rules_modal_form',
            'disable-refocus' => TRUE,
          ],
        ];
      }

      $form['actions']['done'] = [
        '#type' => 'link',
        '#title' => $this->t('Done'),
        '#url' => Url::fromRoute('harmonize.entity_processing_rules'),
        '#attributes' => [
          'class' => ['button'],
        ],
      ];
    }

    return $form;
  }

  /**
   * AJAX callback handler that displays any errors or a success message.
   */
  public function ajax(array &$form, FormStateInterface $form_state) : array {
    return $form;
  }

  /**
   * AJAX callback handler that displays any errors or a success message.
   */
  public function addBundlesSelection(array &$form, FormStateInterface $form_state) : AjaxResponse {
    /**
     * Services.
     *
     * @var EntityTypeBundleInfoInterface $entityTypeBundleInfo
     * @TODO - We don't want to DI because it has weird results with AJAX. Find a solution to this.
     */
    $entityTypeBundleInfo = Drupal::service('entity_type.bundle.info');

    // Get the selected entity type.
    $entityType = $form_state->getValue('entity_type');

    $bundleInfos = $entityTypeBundleInfo->getBundleInfo($entityType);
    if (!empty($bundleInfos)) {
      $bundleOptions = [
        '' => $this->t('--'),
      ];
      foreach ($bundleInfos as $id => $bundleInfo) {
        $bundleOptions[$id] = $bundleInfo['label'];
      }
      asort($bundleOptions);
      $form['bundle']['#options'] = $bundleOptions;
    } else {
      $form['bundle']['#value'] = null;
    }

    $response = new AjaxResponse();
    $response->addCommand(new ReplaceCommand('#js-ajax-bundle-wrapper', $form['bundle']));

    $form_state->setRebuild();

    return $response;
  }

  /**
   * AJAX callback handler that adds a field to a ruleset.
   */
  public function addField(array &$form, FormStateInterface $form_state) : array {
    /**
     * Services.
     * @TODO - We don't want to DI because it has weird results with AJAX. Find a solution to this.
     */
    $messenger = Drupal::messenger();

    // Get the button that triggered the ajax call.
    $button = $form_state->getTriggeringElement();
    if (!$button) {
      return $form;
    }

    // Get our data from the clicked button.
    /* @var FieldConfigInterface $field */
    $field = $button['#field'];
    $entityType = $button['#entityType'];
    $bundle = $button['#bundle'];

    // Translation Args.
    $args['%label'] = $field->getLabel();

    // Get current field config if any.
    $rules = $this->config(HarmonizeConfig::ENTITY_RULES)->get(HarmonizeConfig::ENTITY_RULES__RULES);

    // Check if the rule is set already.
    if ($bundle) {
      if (isset($rules[$entityType]['bundles'][$bundle]['rules'][$field->getName()])) {
        $messenger->addError($this->t('Field %label is already defined in the rules.', $args));
        return $form;
      }
    } else {
      if (isset($rules[$entityType]['rules'][$field->getName()])) {
        $messenger->addError($this->t('Field %label is already defined in the rules.', $args));
        return $form;
      }
    }

    // Now set default values for the new field rule.
    if ($bundle) {
      $rules[$entityType]['bundles'][$bundle]['rules'][$field->getName()] = [];
    } else {
      $rules[$entityType]['rules'][$field->getName()] = [];
    }

    // If we reach here then rules aren't already set, so we set them in config.
    $this->config(HarmonizeConfig::ENTITY_RULES)
      ->set(HarmonizeConfig::ENTITY_RULES__RULES, $rules)
      ->save();

    // A nice little message before we exit.
    $messenger->addStatus($this->t('Field %label was added.', $args));
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {}

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    // Step 1 submission.
    if ($this->step === 1) {
      // Simply increment the step and move forward.
      // At this point, the user simply selected an Entity Type and Bundle if applicable.
      $this->step++;
      $form_state->setRebuild();
    }

    // There are actually no step 2 submissions. We manage the rest without actual submissions.
  }

}
