<?php

namespace Drupal\harmonize\Form;

use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\harmonize\Constants\HarmonizeConfig;

/**
 * Provides a configuration form for the Harmonizer module.
 *
 * @package Drupal\harmonize\Form
 */
class HarmonizeSettingsForm extends ConfigFormBase {

  /**
   * {@inheritdoc}
   */
  public function getFormId() : string {
    return 'harmonize_admin_settings_form';
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames() : array {
    return [
      HarmonizeConfig::MAIN,
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) : array {
    // Set this for better organization of form values later.
    $form['#tree'] = TRUE;

    // Get the current configurations.
    $harmonyFileDiscovery = $this->config(HarmonizeConfig::MAIN)->get(HarmonizeConfig::MAIN__HARMONY_FILE_DISCOVERY);
    $removeFieldUnderscorePrefix = $this->config(HarmonizeConfig::MAIN)->get(HarmonizeConfig::MAIN__REMOVE_FIELD_UNDERSCORE_PREFIX);
    $extensionLoadingAlgorithm = $this->config(HarmonizeConfig::MAIN)->get(HarmonizeConfig::MAIN__LOADING_ALGORITHM);
    $allowedAdminRoutes = $this->config(HarmonizeConfig::MAIN)->get(HarmonizeConfig::MAIN__ALLOWED_ADMIN_ROUTES);
    $allowedAdminRoutes = empty($allowedAdminRoutes) ? [] : implode("\n", $allowedAdminRoutes);
    $automaticallyProcessedEntities = $this->config(HarmonizeConfig::MAIN)->get(HarmonizeConfig::MAIN__AUTO_HARMONIZED_ENTITIES);
    $automaticallyProcessedEntities = empty($automaticallyProcessedEntities) ? [] : implode("\n", $automaticallyProcessedEntities);

    // Enable the use of Harmony.
    $form[HarmonizeConfig::MAIN__HARMONY_FILE_DISCOVERY] = [
      '#type'          => 'checkbox',
      '#title'         => t('Enable Harmony File Discovery'),
      '#description'   => t('Harmony File Discovery allows you to create <strong>.harmony.php</strong> files in your theme to fine-tune preprocessing before it reaches twig templates.'),
      '#default_value' => !empty($harmonyFileDiscovery) ? $harmonyFileDiscovery : FALSE,
    ];

    // Enable the automatic removal "field_" prefix when harmonization fields.
    $form[HarmonizeConfig::MAIN__REMOVE_FIELD_UNDERSCORE_PREFIX] = [
      '#type'          => 'checkbox',
      '#title'         => t('Omit <strong>"field_"</strong> prefix in processed data'),
      '#description'   => t('Check this box to remove the prefix from field names when preprocessing entities.'),
      '#default_value' => !empty($removeFieldUnderscorePrefix) ? $removeFieldUnderscorePrefix : FALSE,
    ];

    // Enable the automatic removal "field_" prefix when harmonization fields.
    $form[HarmonizeConfig::MAIN__LOADING_ALGORITHM] = [
      '#type'          => 'select',
      '#title'         => t('Extension Loading Algorithm'),
      '#description'   => t(
        'This setting touches upon the behaviour when loading objects that alter the preprocessed data. Therefore, <strong>Refiners</strong> & <strong>Harmony Files</strong><br>
              When creating these alteration objects, you have an option to target the data you want to alter via signatures. <em>e.g. <strong>entity.node.basic_page</strong></em><br>
              In the case where multiple alteration objects apply to a single set of data, this algorithm will determine what happens.<br>
              <ul>
                <li><strong>FCFS (First Come First Serve)</strong> : Will only take the most relevant alteration object. <em>e.g. Will take <strong>entity.node.basic_page</strong> over <strong>entity.node</strong> if both exist. Only one alteration object is processed.</em></li>
                <li><strong>SCAN (Elevator)</strong> : Will take all relevant alteration objects. <em>e.g. Will take both <strong>entity.node.basic_page</strong> and <strong>entity.node</strong> if both exist. The smaller signature is processed first, and so forth, so <strong>entity.node</strong> -> <strong>entity.node.basic_page</strong></em></li>
              </ul>
              Refer to the guide for more details on this setting!'
      ),
      '#options'       => [
        'fcfs' => $this->t('FCFS'),
        'scan' => $this->t('SCAN'),
      ],
      '#default_value' => !empty($extensionLoadingAlgorithm) ? $extensionLoadingAlgorithm : 'fcfs',
    ];

    // Configure allowed admin routes.
    $form[HarmonizeConfig::MAIN__AUTO_HARMONIZED_ENTITIES] = [
      '#type'          => 'textarea',
      '#title'         => t('Automatically Processed Entities'),
      '#description'   => t(
        'On each line, enter entity types that should be automatically processed.<br>
              By default, when preprocessing entity reference fields, you must manually preprocess nested entities.<br>
              This default behaviour was put in place to avoid dangerous recursive automatic preprocessing, giving more control to the developers.<br>
              Be mindful of the entity types you set here. Do not automatically preprocess heavy entities.<br><br>
              <em>e.g. <strong>media</strong></em><br>
              <em>e.g. <strong>taxonomy_term</strong></em><br>
              <em>e.g. <strong>paragraph.carousel_slide</strong></em><br>'
      ),
      '#default_value' => !empty($automaticallyProcessedEntities) ? $automaticallyProcessedEntities : '',
    ];

    // Configure allowed admin routes.
    $form[HarmonizeConfig::MAIN__ALLOWED_ADMIN_ROUTES] = [
      '#type'          => 'textarea',
      '#title'         => t('Allowed Admin Routes'),
      '#description'   => t(
        'On each line, enter routes where harmonization should execute when rendering entities.<br>
              Harmonize does not run on admin routes by default since it is intended for the frontend layer.<br>
              That being said, there are cases where you want it to run on the admin side.<br>
              The best example is in the case of the layout builder routes. Those routes are on the admin side, but preview content as it would look on the front-end side, therefore needing preprocessing and theming.<br><br>
              <em>e.g. <strong>layout_builder.add_block</strong></em><br>
              <em>e.g. <strong>layout_builder.move_block</strong></em><br>
              <em>e.g. <strong>layout_builder.remove_block</strong></em><br>'
      ),
      '#default_value' => !empty($allowedAdminRoutes) ? $allowedAdminRoutes : '',
    ];

    // Return the form with all necessary fields.
    return parent::buildForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) : void {
    // Retrieve the configuration.
    $values = $form_state->getValues();

    // Format values.
    $adminRoutes = empty($values[HarmonizeConfig::MAIN__ALLOWED_ADMIN_ROUTES]) ? [] : array_map('trim', explode("\n", $values[HarmonizeConfig::MAIN__ALLOWED_ADMIN_ROUTES]));
    $automaticallyProcessedEntities = empty($values[HarmonizeConfig::MAIN__AUTO_HARMONIZED_ENTITIES]) ? [] : array_map('trim', explode("\n", $values[HarmonizeConfig::MAIN__AUTO_HARMONIZED_ENTITIES]));

    $this->configFactory->getEditable(HarmonizeConfig::MAIN)
      ->set(HarmonizeConfig::MAIN__HARMONY_FILE_DISCOVERY, (bool) $values[HarmonizeConfig::MAIN__HARMONY_FILE_DISCOVERY])
      ->set(HarmonizeConfig::MAIN__REMOVE_FIELD_UNDERSCORE_PREFIX, (bool) $values[HarmonizeConfig::MAIN__REMOVE_FIELD_UNDERSCORE_PREFIX])
      ->set(HarmonizeConfig::MAIN__LOADING_ALGORITHM, $values[HarmonizeConfig::MAIN__LOADING_ALGORITHM])
      ->set(HarmonizeConfig::MAIN__ALLOWED_ADMIN_ROUTES, $adminRoutes)
      ->set(HarmonizeConfig::MAIN__AUTO_HARMONIZED_ENTITIES, $automaticallyProcessedEntities)
      ->save();

    parent::submitForm($form, $form_state);
  }

}
