<?php

namespace Drupal\harmonize\Harmonizer\EntityHarmonizer\EntityFieldHarmonizer;

use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\FieldStorageDefinitionInterface;
use Drupal\harmonize\Harmonizer\Harmonizer;
use Drupal\harmonize\Service\Harmonize;

/**
 * EntityFieldHarmonizer for a given field.
 *
 * This object is used to return clean field data.
 *
 * @property Harmonize $harmonizeService
 *
 * @package Drupal\harmonize\Harmonizer\EntityHarmonizer\EntityFieldHarmonizer
 */
class EntityFieldHarmonizer extends Harmonizer implements EntityFieldHarmonizerInterface {

  /**
   * Field Data definition.
   *
   * Contains all the field's data for the given entity.
   *
   * @var FieldItemListInterface
   */
  protected FieldItemListInterface $fieldData;

  /**
   * Field Name.
   *
   * Contains the name of the field being harmonized.
   *
   * @var string
   */
  protected string $fieldName;

  /**
   * The value(s) of the field.
   *
   * @var mixed
   */
  protected $fieldValues;

  /**
   * Contains the type of field in string format.
   *
   * @var string
   */
  protected string $fieldType;

  /**
   * Contains the field definition declaration.
   *
   * @var FieldDefinitionInterface
   */
  protected FieldDefinitionInterface $fieldDefinition;

  /**
   * Contains the Drupal storage of the field.
   *
   * @var FieldStorageDefinitionInterface
   */
  protected FieldStorageDefinitionInterface $fieldStorage;

  /**
   * Contains the field cardinality.
   *
   * @var int
   */
  protected int $cardinality;

  /**
   * EntityFieldHarmonizer constructor.
   *
   * @param FieldItemListInterface $fieldData
   *   The field data of the field being harmonized.
   * @param Harmonize $harmonizeService
   *   Harmonize module's main service containing Drupal services.
   * @param string $type
   *   The type of harmonizer defined by a string key.
   * @param string $signature
   *   The signature of the harmonizer generated by the factory.
   */
  public function __construct(FieldItemListInterface $fieldData, Harmonize $harmonizeService, string $type, string $signature) {
    parent::__construct($fieldData, $harmonizeService, $type, $signature);
    $this->fieldData = $fieldData;
    $this->fieldName = $fieldData->getName();
    $this->fieldValues = $fieldData->getValue();
    $this->fieldDefinition = $fieldData->getFieldDefinition();
    $this->fieldType = $this->fieldDefinition->getType();
    $this->fieldStorage = $this->fieldDefinition->getFieldStorageDefinition();
    $this->cardinality = $this->fieldStorage->getCardinality();
  }

  /**
   * {@inheritDoc}
   */
  public function getFieldType() : string {
    return $this->fieldType;
  }

  /**
   * {@inheritDoc}
   */
  public function getFieldData() : FieldItemListInterface {
    return $this->fieldData;
  }

  /**
   * {@inheritdoc}
   */
  public function fetchCachedData(): bool {
    // Cache is handled in parent Entity Harmonizers.
    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function isCacheable(): bool {
    // Cache is handled in parent Entity Harmonizers.
    return FALSE;
  }

  /**
   * Formats the data of the field in a clean format.
   *
   * @param array $value
   *   Array of values for the field attached to this harmonizer.
   * @param int $i
   *   The index of the field value.
   *
   * @return mixed
   *   Data of the field for this harmonizer in a clean format.
   */
  protected function process(array $value, int $i) {
    // Normally we get $value['value']. But if that doesn't exist, we just return what Drupal gives us.
    if (!isset($value['value']) || !array_key_exists('value', $value)) {
      return $value;
    }

    return $value['value'];
  }

  /**
   * {@inheritdoc}
   */
  protected function getHarmonizedData() {
    // Check if this is a single-value field.
    if ($this->cardinality === 1) {
      return $this->process($this->fieldValues[0], 0);
    }

    // Array that will hold the field data.
    $data = [];

    // Process all field values.
    foreach ($this->fieldValues as $i => $value) {
      $deltaData = $this->process($value, $i);
      if ($deltaData !== NULL) {
        $data[] = $deltaData;
      }
    }

    return $data;
  }

}
