<?php

namespace Drupal\harmonize\Harmonizer\EntityHarmonizer\EntityFieldHarmonizer;

use Drupal;
use Drupal\Component\Plugin\Exception\PluginException;
use Drupal\harmonize\Constants\HarmonizerFlags;
use Drupal\harmonize\Constants\HarmonizerTypes;
use Drupal\harmonize\Harmonizer\HarmonizerFactory;
use Drupal\harmonize\PluginManager\EntityFieldHarmonizer\EntityFieldHarmonizerPluginManager;

/**
 * Handles the creation of the proper field harmonizer with a given Field.
 *
 * There are many types of Fields in Drupal. Each
 * Field will have their own specific arrangements that will affect the way the
 * data is pre-processed. We will handle these in respective classes.
 *
 * By checking the field type, the Factory builds the appropriate EntityFieldHarmonizer.
 *
 * @package Drupal\harmonize\Harmonizer\EntityHarmonizer\EntityFieldHarmonizer
 */
final class EntityFieldHarmonizerFactory extends HarmonizerFactory {

  /**
   * {@inheritdoc}
   */
  protected static function type() : string {
    return HarmonizerTypes::ENTITY_FIELD;
  }

  /**
   * {@inheritdoc}
   */
  protected static function resolveSignature($object, array $flags = []) : string {
    // Add language.
    $lang = isset($flags[HarmonizerFlags::LANG]) && !empty($flags[HarmonizerFlags::LANG]) ? $flags[HarmonizerFlags::LANG] : Drupal::languageManager()->getCurrentLanguage()->getId();

    // Generate a signature for the harmonizer about to be instantiated.
    // Example: field.email.field_contact_email.
    return self::generateHarmonizerSignature(
      HarmonizerTypes::ENTITY_FIELD,
      $object->getFieldDefinition()->getType(),
      $object->getName(),
      $object->getFieldDefinition()->getTargetEntityTypeId(),
      $object->getFieldDefinition()->getTargetBundle(),
      self::generateFlagsSignatures($flags),
      $object->getEntity()->id(),
      $lang
    );
  }

  /**
   * {@inheritdoc}
   */
  protected static function resolveClass($object) : string {
    // Otherwise, we instantiate it and register it in the factory.
    // Get the type of field.
    $fieldType = $object->getFieldDefinition()->getType();

    // Depending on the type, create the appropriate HarmonizerInterface.
    switch ($fieldType) {
      case 'file':
        $class = FileEntityFieldHarmonizer::class;
        break;

      case 'boolean':
        $class = BooleanEntityFieldHarmonizer::class;
        break;

      case 'daterange':
        $class = DateRangeEntityFieldHarmonizer::class;
        break;

      case 'datetime':
        $class = DateTimeEntityFieldHarmonizer::class;
        break;

      case 'image':
        $class = ImageEntityFieldHarmonizer::class;
        break;

      case 'metatag':
        $class = MetaTagEntityFieldHarmonizer::class;
        break;

      case 'link':
        $class = LinkEntityFieldHarmonizer::class;
        break;

      case 'entity_reference':
        $class = EntityReferenceEntityFieldHarmonizer::class;
        break;

      case 'entity_reference_revisions':
        $class = EntityReferenceRevisionsEntityFieldHarmonizer::class;
        break;

      case 'text_with_summary':
        $class = TextWithSummaryEntityFieldHarmonizer::class;
        break;

      case 'color_field_type':
        $class = ColorEntityFieldHarmonizer::class;
        break;

      case 'address':
        $class = AddressEntityFieldHarmonizer::class;
        break;

      case 'geofield':
        $class = GeofieldEntityFieldHarmonizer::class;
        break;

      case 'physical_measurement':
        $class = PhysicalMeasurementEntityFieldHarmonizer::class;
        break;

      case 'office_hours':
        $class = OfficeHoursEntityFieldHarmonizer::class;
        break;

      default:
        Drupal::logger('harmonize')->warning(t('The field type @fieldType is currently not compatible with Harmonize. Consider requesting the creation of an EntityFieldHarmonizer to cover this field type.', [ '@fieldType' => $fieldType ]));
        $class = EntityFieldHarmonizer::class;
        break;
    }

    return $class;
  }

}
