<?php

namespace Drupal\harmonize\Harmonizer\EntityHarmonizer;

use Drupal\Component\Serialization\Json;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EditorialContentEntityBase;
use Drupal\Core\Entity\EntityInterface;
use Drupal\field\FieldConfigInterface;
use Drupal\harmonize\Constants\HarmonizeConfig;
use Drupal\harmonize\Constants\HarmonizerFlags;
use Drupal\harmonize\Constants\HarmonizerTypes;
use Drupal\harmonize\Entity\EntitySettings;
use Drupal\harmonize\Entity\Style;
use Drupal\harmonize\Harmonizer\EntityHarmonizer\EntityFieldHarmonizer\EntityFieldHarmonizer;
use Drupal\harmonize\Harmonizer\EntityHarmonizer\EntityFieldHarmonizer\EntityFieldHarmonizerFactory;
use Drupal\harmonize\Harmonizer\Harmonizer;
use Drupal\harmonize\Service\Harmonize;
use Exception;

/**
 * Harmonizer for a given entity.
 *
 * This object is used to return clean field data from an Entity.
 *
 * @property Harmonize $harmonizeService
 *
 * @package Drupal\harmonize\Harmonizer\EntityHarmonizer
 */
class EntityHarmonizer extends Harmonizer implements EntityHarmonizerInterface {

  /**
   * Entity attached to the Harmonizer.
   *
   * @var EntityInterface
   */
  protected EntityInterface $entity;

  /**
   * Contains the Entity Type ID.
   *
   * Examples: Node, Paragraph, Media.
   *
   * @var string
   */
  protected string $entityTypeId;

  /**
   * Contains the Entity Bundle, if any.
   *
   * @var string
   */
  protected string $entityBundle;

  /**
   * Array to store configurations for entities loaded from the site.
   *
   * @var array
   */
  protected array $configurations;

  /**
   * Entity Harmonizer base constructor.
   *
   * @param EntityInterface $entity
   *   The entity being harmonized.
   * @param Harmonize $harmonizeService
   *   Harmonize module's main service containing Drupal services.
   * @param string $type
   *   The type of harmonizer defined by a string key.
   * @param string $signature
   *   The signature of the harmonizer generated by the factory.
   */
  public function __construct(
    EntityInterface $entity,
    Harmonize       $harmonizeService,
    string          $type,
    string          $signature
  ) {
    parent::__construct($entity, $harmonizeService, $type, $signature);
    $this->entity = $entity;
    $this->entityTypeId = $entity->getEntityType()->id();
    $this->entityBundle = $entity->bundle();
  }

  /**
   * {@inheritDoc}
   */
  public function getEntityTypeId() : string {
    return $this->entityTypeId;
  }

  /**
   * {@inheritDoc}
   */
  public function getEntityBundle() : string {
    return $this->entityBundle;
  }

  /**
   * {@inheritDoc}
   */
  public function getEntity() : EntityInterface {
    return $this->entity;
  }

  /**
   * {@inheritDoc}
   */
  public function get(string $fieldName) {
    // Get field configurations.
    $fieldConfigs = $this->harmonizeService->helpers->getEntityTypeBundleFieldDefinitions($this->getEntityTypeId(), $this->getEntityBundle());

    // If the field doesn't exist, we simply return null.
    if (!isset($fieldConfigs[$fieldName])) {
      // Try to get value from entity.
      try {
        return $this->entity->get($fieldName)->getValue() ?? NULL;
      } catch (Exception $e) {
        return NULL;
      }
    }

    // Set field config.
    $fieldConfig = $fieldConfigs[$fieldName];

    // Store whether this field is cacheable.
    $fieldIsCacheable = $this->fieldIsCacheable($fieldName, $fieldConfig->getType());

    if ($fieldIsCacheable) {
      $cachedData = $this->fetchCachedFieldData($fieldName);
      if ($cachedData) {
        return $cachedData;
      }
    }

    // Get the field data using the harmonizer.
    $harmonizedFieldData = $this->getHarmonizedFieldData($fieldConfig->getName());

    if ($fieldIsCacheable) {
      if (!empty($harmonizedFieldData)) {
        $this->setCachedFieldData($fieldName, $harmonizedFieldData);
      } else {
        $this->trace(t("Field value is empty for <strong><code>[@field]</code></strong>...Cached data will not be stored", ['@field' => $fieldName]));
      }
    }

    return $harmonizedFieldData;
  }

  /**
   * {@inheritdoc}
   */
  public function getCacheTags() : array {
    // Get the entity's normal cache tags and add tags specific to Harmonize.
    // Even if cache is not enable for this harmonizer, these tags are passed to child field harmonizers.
    return array_merge($this->entity->getCacheTags(), [
      HarmonizerTypes::ENTITY,
      HarmonizerTypes::ENTITY . ':' . $this->entity->getEntityTypeId(),
      HarmonizerTypes::ENTITY . ':' . $this->entity->getEntityTypeId() . ':' . $this->entityBundle,
      HarmonizerTypes::ENTITY . ':' . $this->entity->getEntityTypeId() . ':' . $this->entityBundle . ':' . $this->entity->id(),
    ]);
  }

  /**
   * {@inheritdoc}
   */
  public function fetchCachedData() : bool {
    if ($this->entity instanceof EditorialContentEntityBase) {
      // Cache is handled per field basis in Entity Harmonizers.
      return FALSE;
    }
    return parent::fetchCachedData();
  }

  /**
   * Set cache data.
   */
  protected function setCachedData($data) : void {
    if ($this->entity instanceof EditorialContentEntityBase) {
      // Cache is handled per field basis in Entity Harmonizers.
      return;
    }
    parent::setCachedData($data);
  }

  /**
   * {@inheritdoc}
   */
  protected function getHarmonizedData() {
    // Data array that will be returned.
    $data = [];

    // Trace.
    if ($this->traceable()) {
      if (!Harmonize::$config[HarmonizeConfig::CACHE__ENABLED]) {
        $this->trace(t('Cache disabled in module configurations. Cached data will not be fetched.'));
      }
      if (!$this->isCacheable()) {
        $this->trace(t('This harmonizer type is set as not cacheable. Cached will not be fetched.'));
      }
      if (!$this->cacheEnabled()) {
        $this->trace(t('Cache disabled for this harmonizer. Cached data will not be fetched.'));
      }
    }

    // For now, we only go through with the rest if we're in a ContentEntity.
    // We will properly segment this later on for different types of entities.
    // ...Or maybe we won't! We'll see.
    // @TODO - Consider creating a ContentEntityHarmonizer class and moving most of the functionality here into it.
    // @TODO - After the above, this class should still exist and process simple base field values for an entity.
    if (!($this->entity instanceof ContentEntityInterface)) {
      return $data;
    }

    // First, we want to set the harmonizer to the appropriate language.
    // Translate the entity to the appropriate language if a flag is set.
    // Otherwise, translate it to the current language.
    $this->entity = $this->hasFlag(HarmonizerFlags::LANG) && !empty($this->getFlag(HarmonizerFlags::LANG)) ? $this->harmonizeService->helpers->ensureTranslated($this->entity, $this->getFlag(HarmonizerFlags::LANG)) : $this->harmonizeService->helpers->ensureTranslated($this->entity);

    // Add general information about the entity.
    $data['id'] = $this->getEntity()->id();
    $data['bundle'] = $this->getEntity()->bundle();

    // Get field configurations.
    $fieldConfigs = $this->harmonizeService->helpers->getEntityTypeBundleFieldDefinitions($this->getEntityTypeId(), $this->getEntityBundle());

    // If the FIELDS flag is set...
    if ($this->hasFlag(HarmonizerFlags::FIELDS)) {
      $this->trace(t("Fields flag detected. Only the following fields will be processed: <strong><code>@fields</code></strong>", ['@fields' => Json::encode($this->getFlag(HarmonizerFlags::FIELDS))]));
      // We filter our fields that must be filtered.
      $this->filterFields($fieldConfigs);
    }

    // If the EXCLUDED_FIELDS flag is set...
    if ($this->hasFlag(HarmonizerFlags::EXCLUDED_FIELDS)) {
      $this->trace(t("Field Exclusions flag detected. The following fields will be excluded: <strong><code>@fields</code></strong>", ['@fields' => Json::encode($this->getFlag(HarmonizerFlags::EXCLUDED_FIELDS))]));
      // We filter our fields that must be filtered.
      $this->filterExcludedFields($fieldConfigs);
    }

    // If this entity is being harmonized with a specific view mode...
    if ($this->hasFlag(HarmonizerFlags::VIEW_MODE)) {
      if ($this->harmonizeService->helpers->viewModeExists($this->entityTypeId, $this->getFlag(HarmonizerFlags::VIEW_MODE))) {
        // We filter our fields that must be filtered.
        $this->trace(t("View mode <strong><code>[@mode]</code></strong> detected and identified. Processed fields will be filtered according to the entity's display settings for this view mode", [
          '@mode' => $this->getFlag(HarmonizerFlags::VIEW_MODE)
        ]));
        $this->filterViewModeFields($fieldConfigs);
      } else {
        // We trace.
        $this->trace(t("View mode <strong><code>[@mode]</code></strong> could not be identified in the codebase. The entity will be processed as normal", [
          '@mode' => $this->getFlag(HarmonizerFlags::VIEW_MODE)
        ]));
      }
    }

    // If this entity is being harmonized in a specified style...
    if ($this->hasFlag(HarmonizerFlags::STYLE)) {
      if ($this->harmonizeService->helpers->styleExists($this->getFlag(HarmonizerFlags::STYLE))) {
        // We filter our fields that must be filtered.
        $this->trace(t("Style <strong><code>[@style]</code></strong> detected and identified. Processed fields will be filtered according to the Style's configurations", [
          '@style' => $this->getFlag(HarmonizerFlags::STYLE)
        ]));
        $this->filterStyleFields($fieldConfigs);
      } else {
        // We trace.
        $this->trace(t("Style <strong><code>[@style]</code></strong> could not be identified in the codebase. The entity will be processed as normal", [
          '@style' => $this->getFlag(HarmonizerFlags::STYLE)
        ]));
      }
    }

    // A little bit of tracing. :)
    if (Harmonize::$config[HarmonizeConfig::MAIN__REMOVE_FIELD_UNDERSCORE_PREFIX]) {
      $this->trace(t("The <strong><code>field_</code></strong> prefix will be removed in processed data."));
    }

    // Process all fields for the given entity.
    /* @var FieldConfigInterface $fieldConfig */
    foreach ($fieldConfigs as $fieldConfig) {
      // Get the field name.
      $fieldName = $fieldConfig->getName();

      // If this field is configured to be excluded, skip it in this loop.
      if ($this->fieldIsExcluded($fieldName)) {
        continue;
      }

      // Remove "field_" prefix if we have to.
      if (Harmonize::$config[HarmonizeConfig::MAIN__REMOVE_FIELD_UNDERSCORE_PREFIX]) {
        $harmonizedFieldKey = str_replace('field_', '', $fieldName);
      }
      else {
        $harmonizedFieldKey = $fieldName;
      }

      // Store whether this field is cacheable.
      $fieldIsCacheable = $this->fieldIsCacheable($fieldName, $fieldConfig->getType());

      if ($fieldIsCacheable) {
        $cachedData = $this->fetchCachedFieldData($fieldName);
        if ($cachedData) {
          $this->trace(t("Cached data loaded for <strong><code>[@field]</code></strong>", ['@field' => $fieldName]));
          $data[$harmonizedFieldKey] = $cachedData;
          continue;
        } else {
          $this->trace(t("Cached data not found for <strong><code>[@field]</code></strong>...Proceeding", ['@field' => $fieldName]));
        }
      }

      // Get the field data using the harmonizer.
      $harmonizedFieldData = $this->getHarmonizedFieldData($fieldConfig->getName());

      if ($fieldIsCacheable) {
        if (!empty($harmonizedFieldData)) {
          $this->setCachedFieldData($fieldName, $harmonizedFieldData);
        } else {
          $this->trace(t("Field value is empty for <strong><code>[@field]</code></strong>...Cached data will not be stored", ['@field' => $fieldName]));
        }
      }

      // Set our field data to the array.
      $data[$harmonizedFieldKey] = $harmonizedFieldData;
    }

    return $data;
  }

  /**
   * Harmonize the given Field and return the appropriate data.
   *
   * @param string $fieldName
   *   Field name of the field being harmonized.
   *
   * @return array|string|null
   *   Return the field data in a proper, uniform format.
   */
  protected function getHarmonizedFieldData(string $fieldName) {
    // Get the field data of the entity's field.
    /* @var ContentEntityInterface $entity */
    $entity = $this->getEntity();
    $entityFieldData = $entity->get($fieldName);

    // If the data is empty, return null.
    // @TODO - You can manage this in each EntityFieldHarmonizer by checking if the field is empty there instead.
    // @TODO - This will give more flexibility as it will allow to send different empty values depending on the field.
    if ($entityFieldData->isEmpty()) {
      return [];
    }

    // Build field harmonizer.
    /* @var EntityFieldHarmonizer $fieldHarmonizer */
    $fieldHarmonizer = EntityFieldHarmonizerFactory::build($entityFieldData, $this->getBequeathedFlags(), $this->harmonizeService);

    // Harmonize the field data using the Harmonize module's main service.
    return $fieldHarmonizer->harmonize();
  }

  /**
   * Check if a field is cacheable.
   */
  protected function fieldIsCacheable(string $fieldName, string $fieldType) : bool {
    // Lots of checks. They're self-explanatory.
    if (!Harmonize::$config[HarmonizeConfig::CACHE__ENABLED]) {
      $this->trace(t("Field <strong><code>[@field]</code></strong> is not cacheable. Reason: Cache disabled in Harmonize configurations", ['@field' => $fieldName]));
      return FALSE;
    }
    if (!$this->isCacheable()) {
      $this->trace(t("Field <strong><code>[@field]</code></strong> is not cacheable. Reason: This harmonizer is not cacheable", ['@field' => $fieldName]));
      return FALSE;
    }
    if (!$this->cacheEnabled()) {
      $this->trace(t("Field <strong><code>[@field]</code></strong> is not cacheable. Reason: Cache disabled for this harmonizer", ['@field' => $fieldName]));
      return FALSE;
    }
    if (!$this->fieldCacheIsNotDisabledInRules($fieldName)) {
      return FALSE;
    }
    if ($fieldType === 'entity_reference' || $fieldType === 'entity_reference_revisions') {
      $this->trace(t("Field <strong><code>[@field]</code></strong> is not cacheable. Reason: Entity Reference Fields reference their own harmonizers, which manage their own cache", ['@field' => $fieldName]));
      return FALSE;
    }

    // The field is cacheable if all the checks above pass.
    return TRUE;
  }

  /**
   * Fetch cached data for this harmonizer.
   */
  protected function fetchCachedFieldData(string $field) {
    // Cache actions.
    if (Harmonize::$config[HarmonizeConfig::CACHE__ENABLED] && $this->isCacheable() && $this->cacheEnabled()) {
      $cachedData = $this->harmonizeService->cacheBackend->get($this->getFieldCacheId($field));
      if ($cachedData) {
        $this->data = $cachedData->data;
        return $cachedData->data;
      }
    }

    return NULL;
  }

  /**
   * Set cache data for a field.
   */
  protected function setCachedFieldData(string $field, $data): void {
    if (Harmonize::$config[HarmonizeConfig::CACHE__ENABLED] && $this->isCacheable() && $this->cacheEnabled()) {
      $this->harmonizeService->cacheBackend->set($this->getFieldCacheId($field), $data, CacheBackendInterface::CACHE_PERMANENT, $this->getCacheTags());
      $this->trace(t("Data set in cache with ID <strong><code>[@id]</code></strong>", ['@id' => $this->getFieldCacheId($field)]));
    }
  }

  /**
   * Get the cache ID of a harmonizer's entity field.
   *
   * @param string $field
   *   The field to get a cache ID for.
   *
   * @return string
   */
  protected function getFieldCacheId(string $field): string {
    return $this->getCacheId() . '.field.' . $field;
  }

  /**
   * Check if the given field is excluded for harmonization.
   *
   * @param string $fieldName
   *   Name of the field to check.
   *
   * @return bool
   *   Returns true if the field is excluded. Returns false otherwise.
   */
  protected function fieldIsExcluded(string $fieldName): bool {
    // Get entity rules configurations.
    $rules = Harmonize::$config[HarmonizeConfig::ENTITY_RULES__RULES];

    // Validations.
    if ($this->isExcludedInRules($fieldName, $rules)) {
      $this->trace(t("Field <strong><code>[@field]</code></strong> has been excluded via Entity Rules (Config).", ['@field' => $fieldName]));
      return TRUE;
    }

    // Get entity rules configurations.
    $themeRules = $this->harmonizeService->helpers->getHarmonyThemeRulesConfig($this->getFlag(HarmonizerFlags::THEME) ?? NULL);

    // Validations.
    if ($this->isExcludedInRules($fieldName, $themeRules)) {
      $this->trace(t("Field <strong><code>[@field]</code></strong> has been excluded via Entity Rules (Theme).", ['@field' => $fieldName]));
      return TRUE;
    }

    // If all checks fail, return false.
    return FALSE;
  }

  /**
   * Check if the given field has cache disabled in rules.
   *
   * @param string $fieldName
   *   Name of the field to check.
   *
   * @return bool
   *   Returns true if the field cache is disabled. Returns false otherwise.
   */
  protected function fieldCacheIsNotDisabledInRules(string $fieldName): bool {
    // Get our rule configurations.
    $rules = Harmonize::$config[HarmonizeConfig::ENTITY_RULES__RULES];

    // Not excluded if configurations for this entity type do not exist.
    $bundleRules = $rules[$this->getEntityTypeId()]['bundles'][$this->getEntityBundle()] ?? NULL;

    // Check if the field is set in the rules.
    if ($bundleRules && array_key_exists($fieldName, $bundleRules['rules'] ?? [])) {
      if (isset($bundleRules['rules'][$fieldName]['disabled_cache']) && $bundleRules['rules'][$fieldName]['disabled_cache']) {
        $this->trace(t("Field <strong><code>[@field]</code></strong> is not cacheable. Reason: Cache disabled in Entity Rules (Config)", ['@field' => $fieldName]));
        return FALSE;
      }
    }

    // Check entity rules.
    $entityRules = $rules[$this->getEntityTypeId()] ?? NULL;

    // Check if the field is set in the rules.
    if ($entityRules && array_key_exists($fieldName, $entityRules['rules'] ?? [])) {
      if (isset($entityRules['rules'][$fieldName]['disabled_cache']) && $entityRules['rules'][$fieldName]['disabled_cache']) {
        $this->trace(t("Field <strong><code>[@field]</code></strong> is not cacheable. Reason: Cache disabled in Entity Rules (Config)", ['@field' => $fieldName]));
        return FALSE;
      }
    }

    // Get harmony configurations from the theme if any.
    $rules = $this->harmonizeService->helpers->getHarmonyThemeRulesConfig($this->getFlag(HarmonizerFlags::THEME) ?? NULL);

    // Not excluded if configurations for this entity type do not exist.
    $bundleRules = $rules[$this->getEntityTypeId()]['bundles'][$this->getEntityBundle()] ?? NULL;

    // Check if the field is set in the rules.
    if ($bundleRules && array_key_exists($fieldName, $bundleRules['rules'] ?? [])) {
      if (isset($bundleRules['rules'][$fieldName]['disabled_cache']) && $bundleRules['rules'][$fieldName]['disabled_cache']) {
        $this->trace(t("Field <strong><code>[@field]</code></strong> is not cacheable. Reason: Cache disabled in Entity Rules (Theme)", ['@field' => $fieldName]));
        return FALSE;
      }
    }

    // Check entity rules.
    $entityRules = $rules[$this->getEntityTypeId()] ?? NULL;

    // Check if the field is set in the rules.
    if ($entityRules && array_key_exists($fieldName, $entityRules['rules'] ?? [])) {
      if (isset($entityRules['rules'][$fieldName]['disabled_cache']) && $entityRules['rules'][$fieldName]['disabled_cache']) {
        $this->trace(t("Field <strong><code>[@field]</code></strong> is not cacheable. Reason: Cache disabled in Entity Rules (Theme)", ['@field' => $fieldName]));
        return FALSE;
      }
    }

    // At this point if we haven't returned yet, the field is not auto harmonized.
    return TRUE;
  }

  /**
   * Validate whether a field is excluded via an array of rules.
   *
   * @param string $fieldName
   *   The field machine name.
   * @param array|null $rules
   *   The array of rules, either populated by config or a theme declaration file.
   *
   * @return bool
   *   Returns TRUE if the field is excluded, returns FALSE otherwise.
   */
  protected function isExcludedInRules(string $fieldName, ?array $rules) : bool {
    // Just return false if rules is empty.
    if (empty($rules)) {
      return FALSE;
    }

    // Check entity rules.
    $entityRules = $rules[$this->getEntityTypeId()] ?? NULL;

    // Check if the field is set in the rules.
    if ($entityRules && array_key_exists($fieldName, $entityRules['rules'] ?? [])) {
      if (isset($entityRules['rules'][$fieldName]['excluded']) && $entityRules['rules'][$fieldName]['excluded']) {
        return TRUE;
      }
    }

    // Not excluded if configurations for this entity type do not exist.
    $bundleRules = $rules[$this->getEntityTypeId()]['bundles'][$this->getEntityBundle()] ?? NULL;

    // Check if the field is set in the rules.
    if ($bundleRules && array_key_exists($fieldName, $bundleRules['rules'] ?? [])) {
      if (isset($bundleRules['rules'][$fieldName]['excluded']) && $bundleRules['rules'][$fieldName]['excluded']) {
        return TRUE;
      }
    }

    return FALSE;
  }

  /**
   * Filter fields that are to be harmonized.
   *
   * @param $field_config_array
   *   The field configuration array.
   */
  protected function filterFields(&$field_config_array) {
    $field_config_array = array_filter($field_config_array, function($fieldMachineName) {
      $check = in_array($fieldMachineName, $this->flags[HarmonizerFlags::FIELDS], TRUE);
      if (!$check) {
        $this->trace(t("Field <strong><code>[@field]</code></strong> filtered by Fields Flag.", ['@field' => $fieldMachineName]));
      }
      return $check;
    }, ARRAY_FILTER_USE_KEY);
  }

  /**
   * Filter fields that are to be excluded.
   *
   * @param $field_config_array
   *   The field configuration array.
   */
  protected function filterExcludedFields(&$field_config_array) {
    $field_config_array = array_filter($field_config_array, function($fieldMachineName) {
      $check = !in_array($fieldMachineName, $this->flags[HarmonizerFlags::EXCLUDED_FIELDS], TRUE);
      if (!$check) {
        $this->trace(t("Field <strong><code>[@field]</code></strong> filtered by Field Exclusions Flag.", ['@field' => $fieldMachineName]));
      }
      return $check;
    }, ARRAY_FILTER_USE_KEY);
  }

  /**
   * Filter fields that are to be harmonized for a given view mode.
   *
   * @param $fieldConfigArray
   *   The field configuration array.
   */
  protected function filterViewModeFields(array &$fieldConfigArray) {
    // At this point we already did the view mode check.
    // Array to store our fields to exclude.
    $fieldsToExclude = [];

    // Now we try to load entity settings.
    $settings = EntitySettings::load($this->entityTypeId . '.' . $this->entityBundle);

    // If the settings are empty, we can stop here.
    if (empty($settings)) {
      $this->trace(t("No display settings found for <strong><code>[@entityType:@bundle]</code></strong>. Fields will not be filtered per the defined View Mode"));
      return;
    }

    // Load display settings.
    $displaySettings = $settings->getDisplaySettings();

    // Check if display settings exist.
    if (empty($displaySettings)) {
      $this->trace(t("Entity Settings exist for <strong><code>[@entityType:@bundle]</code></strong>, but no display settings were found. Fields will not be filtered per the defined View Mode", [
        '@entityType' => $this->entityTypeId,
        '@bundle' => $this->entityBundle,
      ]));
      return;
    }

    // Check if we have display settings for the defined view mode.
    if (!isset($displaySettings[$this->getFlag(HarmonizerFlags::VIEW_MODE)])) {
      $this->trace(t("There are no display settings defined for @viewMode in the Entity Settings. An attempt will be made to find default display settings", [
        '@viewMode' => $this->getFlag(HarmonizerFlags::VIEW_MODE),
      ]));

      // Check if we have default display settings.
      if (!isset($displaySettings['default'])) {
        $this->trace(t("There are no default display settings defined in the Entity Settings. Processed fields will not be filtered", [
          '@viewMode' => $this->getFlag(HarmonizerFlags::VIEW_MODE),
        ]));
      } else {
        // Populate our fields array with default settings configurations.
        foreach ($displaySettings['default'] as $fieldName => $included) {
          if (!$included) {
            $fieldsToExclude[] = $fieldName;
          }
        }

        // Filter our array.
        // We used to do an empty check on the $fields array here. Now we don't.
        // This means that if a style has no fields declared, no fields will be harmonized.
        // This is intended for now. We'll see if we change this later.
        $fieldConfigArray = array_filter($fieldConfigArray, function($fieldMachineName) use($fieldsToExclude) {
          $check = !in_array($fieldMachineName, $fieldsToExclude, TRUE);
          if (!$check) {
            $this->trace(t("Field <strong><code>[@field]</code></strong> excluded by default View Mode settings. The following fields are included: @fields", [
              '@field'  => $fieldMachineName,
              '@mode'  => $this->getFlag(HarmonizerFlags::VIEW_MODE),
              '@fields' => Json::encode($fieldsToExclude),
            ]));
          }
          return $check;
        }, ARRAY_FILTER_USE_KEY);
      }
      return;
    }

    // At this point we can finally populate our fields array.
    foreach ($displaySettings[$this->getFlag(HarmonizerFlags::VIEW_MODE)] as $fieldName => $included) {
      if (!$included) {
        $fieldsToExclude[] = $fieldName;
      }
    }

    // Filter our array.
    // We used to do an empty check on the $fields array here. Now we don't.
    // This means that if a style has no fields declared, no fields will be harmonized.
    // This is intended for now. We'll see if we change this later.
    $fieldConfigArray = array_filter($fieldConfigArray, function($fieldMachineName) use($fieldsToExclude) {
      $check = !in_array($fieldMachineName, $fieldsToExclude, TRUE);
      if (!$check) {
        $this->trace(t("Field <strong><code>[@field]</code></strong> excluded by View Mode: <strong><code>[@mode]</code></strong>. The View Mode only includes the following fields: @fields", [
          '@field'  => $fieldMachineName,
          '@mode'  => $this->getFlag(HarmonizerFlags::VIEW_MODE),
          '@fields' => Json::encode($fieldsToExclude),
        ]));
      }
      return $check;
    }, ARRAY_FILTER_USE_KEY);
  }

  /**
   * Filter fields that are to be harmonized for a given style.
   *
   * @param $fieldConfigArray
   *   The field configuration array.
   */
  protected function filterStyleFields(array &$fieldConfigArray) {
    // Check if the style exists first.
    // Array to store our fields to filter.
    $fields = [];

    // Get style from flag.
    $styleName = $this->getFlag(HarmonizerFlags::STYLE);

    // Load the appropriate style.
    $style = Style::load($styleName);

    // If the style is properly loaded we continue.
    if (!empty($style)) {
      // Merge fields declared in style.
      $fields = array_merge($fields, $style->getFields());
    }

    // Load Theme configurations.
    $harmonyThemeConfig = $this->harmonizeService->helpers->getHarmonyThemeStylesConfig();

    // If we have a style of this name declared in the theme, we merge.
    // @TODO - Don't we want theme styles to take precedence? Review this.
    if (isset($harmonyThemeConfig['styles'][$styleName])) {
      $fields = array_merge($fields, $harmonyThemeConfig['styles'][$styleName]['fields'] ?? []);
    }

    // Filter our array.
    // We used to do an empty check on the $fields array here. Now we don't.
    // This means that if a style has no fields declared, no fields will be harmonized.
    // This is intended for now. We'll see if we change this later.
    $fieldConfigArray = array_filter($fieldConfigArray, function($fieldMachineName) use($fields) {
      $check = in_array($fieldMachineName, $fields, TRUE);
      if (!$check) {
        $this->trace(t("Field <strong><code>[@field]</code></strong> excluded by Style: <strong><code>[@style]</code></strong>. The Style only includes the following fields: @fields", [
          '@field'  => $fieldMachineName,
          '@style'  => $this->getFlag(HarmonizerFlags::STYLE),
          '@fields' => Json::encode($fields),
        ]));
      }
      return $check;
    }, ARRAY_FILTER_USE_KEY);
  }

}
