<?php

namespace Drupal\harmonize\Harmonizer\EntityHarmonizer;

use Drupal;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\harmonize\Constants\HarmonizerFlags;
use Drupal\harmonize\Constants\HarmonizerTypes;
use Drupal\harmonize\Harmonizer\HarmonizerFactory;

/**
 * Handles the creation of the proper harmonizer for a given Entity.
 *
 * There are many types of Entities in Drupal. Each Entity may
 * have their own specific arrangements that will affect the way the data is
 * preprocessed. We will handle these in respective classes.
 *
 * By checking the entity type, the Factory builds the appropriate Harmonizer.
 *
 * @package Drupal\harmonize\Harmonizer\EntityHarmonizer\EntityHarmonizer
 */
final class EntityHarmonizerFactory extends HarmonizerFactory {

  /**
   * {@inheritdoc}
   */
  protected static function type() : string {
    return HarmonizerTypes::ENTITY;
  }

  /**
   * {@inheritdoc}
   */
  public static function resolveSignature($object, array $flags = []) : string {
    // Generate a signature for the harmonizer about to be instantiated.
    // Example: entity.node.basic_page.
    $signature = self::generateHarmonizerSignature(HarmonizerTypes::ENTITY, $object->getEntityTypeId(), $object->bundle(), self::generateFlagsSignatures($flags), $object->id());

    // If we're in a content entity, we append the revision ID to the signature.
    if ($object instanceof ContentEntityInterface) {
      $loadedRevisionId = $object->getLoadedRevisionId();
      if ($loadedRevisionId !== NULL) {
        $signature = self::generateHarmonizerSignature($signature, $object->getLoadedRevisionId());
      }
    }

    // Add language.
    $lang = isset($flags[HarmonizerFlags::LANG]) && !empty($flags[HarmonizerFlags::LANG]) ? $flags[HarmonizerFlags::LANG] : Drupal::languageManager()->getCurrentLanguage()->getId();
    $signature .= '.' . $lang;

    return $signature;
  }

  /**
   * {@inheritdoc}
   */
  protected static function resolveClass($object) : string {
    // Get the type of entity.
    // i.e. "node", "paragraph", "media", etc.
    $entity_type = $object->getEntityTypeId();

    // Depending on the type, create the appropriate "Harmonizer".
    switch ($entity_type) {
      case 'node':
        $class = NodeEntityHarmonizer::class;
        break;

      case 'file':
        $class = FileEntityHarmonizer::class;
        break;

      case 'media':
        $class = MediaEntityHarmonizer::class;
        break;

      case 'paragraph':
        $class = ParagraphEntityHarmonizer::class;
        break;

      case 'taxonomy_term':
        $class = TaxonomyTermEntityHarmonizer::class;
        break;

      case 'menu_link_content':
        $class = MenuLinkContentEntityHarmonizer::class;
        break;

      case 'image_style':
        $class = ImageStyleEntityHarmonizer::class;
        break;

      default:
        $class = EntityHarmonizer::class;
        break;
    }

    return $class;
  }

}
