<?php

namespace Drupal\harmonize\Harmonizer;

use Drupal;
use Drupal\Component\Serialization\Json;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\DependencyInjection\DependencySerializationTrait;
use Drupal\harmonize\Constants\HarmonizationEvents;
use Drupal\harmonize\Constants\HarmonizeConfig;
use Drupal\harmonize\Constants\HarmonizerFlags;
use Drupal\harmonize\Event\HarmonizationEvent;
use Drupal\harmonize\Event\HarmonizationEventFactory;
use Drupal\harmonize\Harmonizer\EntityHarmonizer\EntityHarmonizer;
use Drupal\harmonize\Service\Harmonize;

/**
 * Provides a base Class for all Harmonizers.
 *
 * @package Drupal\harmonize\Harmonizer
 */
abstract class Harmonizer implements HarmonizerInterface {

  // Used to avoid problem where database serialization isn't possible.
  use DependencySerializationTrait;

  /**
   * Harmonize Service.
   *
   * @var Harmonize
   */
  protected Harmonize $harmonizeService;

  /**
   * Data obtained from the harmonizer after all pre-processing is complete.
   *
   * @var mixed
   */
  protected $data;

  /**
   * The object being harmonized.
   *
   * The type-hint is "mixed" since this could be anything out of Drupal's many objects.
   *
   * @var mixed
   */
  protected $object;

  /**
   * Set the type of harmonizer this is.
   *
   * @var string
   */
  protected string $type;


  /**
   * Harmonizer signature, defined by the object being harmonized.
   *
   * @var string
   */
  protected string $signature;

  /**
   * Flags that can be set on a harmonizer to alter preprocessing behavior.
   *
   * For a collection of flags that can be set and their effects, check out the HarmonizerFlags class.
   *
   * @var array
   *
   * @see \Drupal\harmonize\Constants\HarmonizerFlags
   */
  protected array $flags;

  /**
   * The trace is an array storing the steps in the harmonization process and information on what's happening.
   *
   * Very useful for debugging the module's very intricate functionalities.
   *
   * @var mixed
   */
  protected $trace = [];

  /**
   * Parent Harmonizer that called this harmonizer.
   *
   * If a harmonizer is called in the field of an existing one, then it is considered the child of the existing
   * harmonizer.
   *
   * @var HarmonizerInterface
   */
  protected HarmonizerInterface $parentHarmonizer;

  /**
   * Harmonizer constructor.
   *
   * @param mixed $object
   *   The object being harmonized.
   * @param Harmonize $harmonizeService
   *   Harmonize module's main service containing Drupal services.
   * @param string $type
   *   The type of harmonizer defined by a string key.
   * @param string $signature
   *   The signature of the harmonizer generated by the factory.
   */
  public function __construct($object, Harmonize $harmonizeService, string $type, string $signature, array $flags = []) {
    $this->object = $object;
    $this->harmonizeService = $harmonizeService;
    $this->type = $type;
    $this->signature = $signature;
    $this->flags = $flags;
  }

  /**
   * Preprocess all fields in this harmonizer and return the data.
   *
   * @param array $flags
   *   Optionally set flags on the fly to use in your Harmonizer.
   *
   * @return array|string|null
   *   The data of the harmonizer if any.
   */
  public function harmonize(array $flags = []) {
    $this->trace(t("Starting harmonization: <strong><code>[@signature]</code></strong>", ['@signature' => $this->getSignature()]));
    // Check if cache was set for this harmonizer.
    if (!empty($this->data)) {
      $this->trace(t("Data already set for this harmonizer...No processing needed"), TRUE);
      return $this->data;
    }

    // Merge the flags provided with the flags already set.
    if ($flags) {
      $this->flags = array_merge($this->flags, $flags);
    }

    // Show flags in trace.
    if (!empty($this->flags) && $this->traceable()) {
      $this->trace(t("Flags have been specified: <strong><code>@flags</code></strong>", ['@flags' => Json::encode($this->flags)]));
    }

    // Attempt to fetch data from cache.
    if (!$this->fetchCachedData()) {
      if ($this->traceable() && !($this instanceof EntityHarmonizer)) {
        $this->trace(t("Data has not been obtained in cache...Beginning data processing"));
      }
      $this->data = $this->getHarmonizedData();
      $this->setCachedData($this->data);
    }

    // Dispatch all Harmonization events.
    $this->trace(t("Initial data processed...Dispatching Harmonization Events"), TRUE);
    $this->dispatchHarmonizationEvents();

    // Return the data.
    $this->trace(t("Harmonization process completed successfully."), TRUE);
    return $this->data;
  }

  /**
   * {@inheritdoc}
   */
  public function getCacheTags() : array {
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function isCacheable() : bool {
    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function cacheEnabled() : bool {
    $signatureParts = $this->getSignatureParts(TRUE);
    foreach ($signatureParts as $part) {
      if (in_array($part, Harmonize::$config[HarmonizeConfig::CACHE__EXCLUSIONS], TRUE)) {
        return FALSE;
      }
    }
    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function getData() {
    return $this->data;
  }

  /**
   * {@inheritdoc}
   */
  public function setData($data) : void {
    $this->data = $data;
  }

  /**
   * {@inheritdoc}
   */
  public function getObject() {
    return $this->object;
  }

  /**
   * {@inheritdoc}
   */
  public function getType() : string {
    return $this->type;
  }

  /**
   * {@inheritdoc}
   */
  public function getSignature() : string {
    return $this->signature;
  }

  /**
   * {@inheritdoc}
   */
  public function setSignature(string $signature) : void {
    $this->signature = $signature;
  }

  /**
   * {@inheritdoc}
   */
  public function getSignatureParts(bool $incremental_parts = FALSE) {
    // Explode into array of elements, using "." as a delimiter.
    // Example: "entity.node" becomes ['entity', 'node'].
    $parts = explode('.', $this->signature);

    // If the incremental parts flag isn't set, return the exploded array.
    if (!$incremental_parts) {
      return $parts;
    }

    // Otherwise, we'll create the desired incremental format.
    // @see this function's documentation for more information.
    $result = [];

    foreach ($parts as $i => $part) {
      $needle = $part;

      if ($i !== 0) {
        $needle = '';
        for ($c = 0; $c <= $i; $c++) {
          $needle .= $parts[$c];
          if ($c !== $i) {
            $needle .= '.';
          }
        }
      }
      $result[] = $needle;
    }

    // Return the formatted result.
    return $result;
  }

  /**
   * {@inheritdoc}
   */
  public function getFlag(string $key) {
    return $this->flags[$key] ?? NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function getFlags() : array {
    return $this->flags ?? [];
  }

  /**
   * {@inheritdoc}
   */
  public function hasFlag(string $key) : bool {
    return array_key_exists($key, $this->flags) && !empty($this->flags[$key]);
  }

  /**
   * {@inheritdoc}
   */
  public function setFlag(string $key, $value) : void {
    $this->flags[$key] = $value;
  }

  /**
   * {@inheritdoc}
   */
  public function setFlags(array $flags) : void {
    $this->flags = $flags;
  }

  /**
   * {@inheritdoc}
   */
  public function getBequeathedFlags() : array {
    $flags = $this->flags ?? [];
    if (!empty($this->flags)) {
      $flags = [
        HarmonizerFlags::LANG => $this->flags[HarmonizerFlags::LANG] ?? '',
        HarmonizerFlags::THEME => $this->flags[HarmonizerFlags::THEME] ?? '',
        HarmonizerFlags::PARENT => $this,
      ];
    }
    return $flags;
  }

  /**
   * {@inheritdoc}
   */
  public function getTrace() : array {
    return $this->trace;
  }

  /**
   * {@inheritdoc}
   */
  public function trace(string $trace, bool $logData = FALSE) : void {
    if (!$this->traceable()) return;
    $this->trace[] = [
      'trace' => $trace,
      'data' => $logData ? $this->data : NULL,
    ];
  }

  /**
   * Add a Trace step in the Harmonizer.
   *
   * @return bool
   *   Returns TRUE if tracing is activated, returns FALSE otherwise.
   */
  public function traceable() : bool {
    return $this->hasFlag(HarmonizerFlags::TRACE) && $this->getFlag(HarmonizerFlags::TRACE);
  }

  /**
   * {@inheritdoc}
   */
  public function getParentHarmonizer() : ?HarmonizerInterface {
    return $this->parentHarmonizer;
  }

  /**
   * {@inheritdoc}
   */
  public function setParentHarmonizer(HarmonizerInterface $harmonizer) : void {
    $this->parentHarmonizer = $harmonizer;
  }

  /**
   * Get the initially harmonized data for the given object.
   *
   * This quite simply loops through the objects properties, fields or
   * other data accordingly, and outputs the data in clean array format.
   *
   * @return mixed
   *   The data properly formatted from the object.
   */
  abstract protected function getHarmonizedData();

  /**
   * Fetch cached data for this harmonizer.
   */
  protected function fetchCachedData() : bool {
    // Cache actions.
    if (Harmonize::$config[HarmonizeConfig::CACHE__ENABLED] && $this->isCacheable() && $this->cacheEnabled()) {
      $cachedData = $this->harmonizeService->cacheBackend->get($this->getCacheId());
      if ($cachedData) {
        $this->data = $cachedData->data;
        return TRUE;
      }
    }

    // Trace.
    if ($this->traceable()) {
      if (!Harmonize::$config[HarmonizeConfig::CACHE__ENABLED]) {
        $this->trace(t('Cache disabled in module configurations. Cached data could not be fetched.'));
      }
      if (!$this->isCacheable()) {
        $this->trace(t('This harmonizer type is set as not cacheable. Cached data could not be fetched.'));
      }
      if (!$this->cacheEnabled()) {
        $this->trace(t('Cache disabled for this harmonizer. Cached data could not be fetched.'));
      }
    }
    return FALSE;
  }

  /**
   * Set cache data.
   */
  protected function setCachedData($data) : void {
    if (Harmonize::$config[HarmonizeConfig::CACHE__ENABLED] && $this->isCacheable()  && $this->cacheEnabled()) {
      $this->harmonizeService->cacheBackend->set($this->getCacheId(), $data, CacheBackendInterface::CACHE_PERMANENT, $this->getCacheTags());
      $this->trace(t("Data set in cache with ID <strong><code>[@id]</code></strong>", ['@id' => $this->getCacheId()]));
    }
  }

  /**
   * Get the cache ID of the harmonizer.
   *
   * @return string
   */
  protected function getCacheId() : string {
    // The cache ID is simply the signature for now.
    // We add the language & theme too.
    return $this->getSignature() .
      '.' . $this->getFlag(HarmonizerFlags::THEME) ?? Drupal::theme()->getActiveTheme()->getName() .
      '.' . $this->getFlag(HarmonizerFlags::LANG) ?? Drupal::languageManager()->getCurrentLanguage()->getId();
  }

  /**
   * Dispatch events related to this Harmonizer.
   *
   * Refiners & Harmony Files react to the events dispatched here to alter preprocessed data.
   *
   * This also allows other modules to react on Harmonization events and alter data.
   */
  private function dispatchHarmonizationEvents() : void {
    // Dispatch the core event.
    $this->harmonizeService->eventDispatcher->dispatch(new HarmonizationEvent($this), HarmonizationEvents::CORE);

    // Build info on the event to dispatch for this Harmonizer.
    $eventInfo = HarmonizationEventFactory::build($this);

    // Dispatch the Harmonizer Type specific event.
    $this->harmonizeService->eventDispatcher->dispatch($eventInfo['event_object'], $eventInfo['event_name']);
  }

}
