<?php

namespace Drupal\harmonize\Service\FormTuner;

use Drupal\Core\Config\Entity\ConfigEntityBundleBase;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Form\FormStateInterface;

/**
 * Provides a form tuner for Entity Type Edit forms.
 *
 * @package Drupal\harmonize\Service\FormTuner
 */
final class EntityTypeEditFormTuner extends FormTuner {

  /**
   * ID of this form tuner.
   */
  public const ID = 'harmonize.form_tuner.entity_type_edit';

  /**
   * {@inheritDoc}
   *
   * @var ConfigEntityBundleBase $formEntity
   */
  public function alter(array &$form, FormStateInterface $formState, EntityInterface $formEntity) : void {
    // Get entity type ID and bundle.
    $entityTypeId = $formEntity->getEntityType()->getBundleOf();
    $bundle = $formEntity->id();

    // Attach library.
    $form['#attached']['library'][] = 'harmonize/entity_type_edit_form';

    // Try to get existing configurations.
    $settings = FormTuner::loadEntitySettings($entityTypeId, $bundle);

    // Wrap everything in a fieldset.
    $form['harmonize'] = [
      '#type' => 'details',
      '#title' => t('Harmonize settings'),
      '#collapsed' => FALSE,
      '#collapsible' => TRUE,
      '#tree' => FALSE,
      '#weight' => 10,
      '#group' => 'additional_settings',
      '#attributes' => ['class' => ['harmonize-entity-settings-form']],
    ];

    $form['harmonize']['populate_variables'] = [
      '#type' => 'checkbox',
      '#title' => t('Populate theme variables with harmonized data'),
      '#default_value' => !empty($settings) && $settings->getPopulateVariables(),
      '#description' => t('If this is checked, a variable called <code>harmony</code> containing clean preprocessed data will be added to templates for this content type.
                        <br>If this is unchecked, this can still be accomplished manually through use of the Harmonize service and preprocess functions.'),
      '#attributes' => ['class' => ['harmonize-entity-settings-populate-setting']],
    ];

    $form['harmonize']['display_filtering'] = [
      '#type' => 'checkbox',
      '#title' => t('Enable display filtering'),
      '#default_value' => !empty($settings) && $settings->getAllowDisplayFiltering(),
      '#description' => t('If this is checked, you will be able to adjust the fields that are preprocessed in the display settings.<br>
                                This will allow you to fine-tune which fields are sent to templates per display type.'),
      '#attributes' => ['class' => ['harmonize-entity-settings-display-filtering-setting']],
    ];

    // Attach our submit handler.
    $form['actions']['submit']['#submit'][] = get_class() . '::submitHandler';
  }

  /**
   * Custom submit handler.
   */
  public static function submitHandler(array $form, FormStateInterface $formState) {
    // Get stuff we need before proceeding.
    $formObject = $formState->getFormObject();
    $formEntity = $formObject->getEntity();
    $entityTypeId = $formEntity->getEntityType()->getBundleOf();
    $bundle = $formEntity->id();

    // Try to get existing configurations.
    // If we can't find existing ones, we create one on the fly.
    $settings = FormTuner::loadEntitySettings($entityTypeId, $bundle, TRUE);

    // Now we get the form values we need and add them to the entity.
    $populateVariables = $formState->getValue('populate_variables');
    $displayFiltering = $formState->getValue('display_filtering');

    // Set them in entity and save.
    $settings->setPopulateVariables($populateVariables);
    $settings->setAllowDisplayFiltering($displayFiltering);
    $settings->save();
  }

}
