<?php

namespace Drupal\harmonize\Service\FormTuner;

use Drupal\Core\Entity\Display\EntityViewDisplayInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\harmonize\Entity\EntitySettings;

/**
 * Provides a form tuner for Entity Type Edit forms.
 *
 * @package Drupal\harmonize\Service\FormTuner
 */
final class EntityTypeViewDisplayFormTuner extends FormTuner {

  /**
   * ID of this form tuner.
   */
  public const ID = 'harmonize.form_tuner.entity_type_view_display';

  /**
   * {@inheritDoc}
   *
   * @var EntityViewDisplayInterface $formEntity
   */
  public function alter(array &$form, FormStateInterface $formState, EntityInterface $formEntity) : void {
    // Get entity type ID and bundle.
    $entityTypeId = $formEntity->getTargetEntityTypeId();
    $bundle = $formEntity->getTargetBundle();

    // Our goal here is to modify the table to include additional options.
    // First, we check if we have Harmonize settings.
    $settings = FormTuner::loadEntitySettings($entityTypeId, $bundle);

    // If settings aren't specified, or we don't have display filtering enabled, we can stop here.
    if (empty($settings) || !$settings->getAllowDisplayFiltering()) {
      return;
    }

    // Attach library.
    $form['#attached']['library'][] = 'harmonize/entity_type_view_display_form';

    // Get the display mode.
    $displayMode = $formEntity->getMode();

    // Get display settings configurations.
    $displaySettings = $settings->getDisplaySettings();

    // Get the table
    $table = &$form['fields'];

    // Add header for preprocess inclusion.
    $table['#header'][] = t('Harmonize');

    // Loop into our rows and add the checkbox.
    foreach ($table as $fieldName => $rowRenderArray) {
      // We don't act on render array keys.
      if (str_starts_with($fieldName, '#')) continue;

      // If the field doesn't start with 'field_', we skip.
      // We take 'body' though. We like 'body'.
      if ($fieldName !== 'body' && !str_starts_with($fieldName, 'field_')) continue;

      // Add harmonize checkbox.
      $table[$fieldName]['harmonize'] = [
        '#type' => 'checkbox',
        '#default_value' => $displaySettings[$displayMode][$fieldName] ?? TRUE,
        '#attributes' => ['class' => ['harmonize-entity-settings-harmonize-checkbox']],
        '#wrapper_attributes' => ['class' => ['harmonize-entity-settings-harmonize-checkbox-wrapper']],
      ];
    }

    // Attach our submit handler.
    $form['actions']['submit']['#submit'][] = get_class() . '::submitHandler';
  }

  /**
   * Custom submit handler.
   */
  public static function submitHandler(array $form, FormStateInterface $formState) {
    // Get entity type ID and bundle.
    $formEntity = $formState->getFormObject()->getEntity();
    $entityTypeId = $formEntity->getTargetEntityTypeId();
    $bundle = $formEntity->getTargetBundle();

    // Load our settings entity.
    $settings = FormTuner::loadEntitySettings($entityTypeId, $bundle);

    // Defensive checks. Might be overkill but whatever.
    if (empty($settings)) {
      return;
    }

    // Load display settings.
    $displaySettings = $settings->getDisplaySettings();

    // Get the display mode.
    $mode = $formEntity->getMode();

    // Get our harmonize values.
    $fields = $formState->getValue('fields');

    // Loop through our fields and set fields.
    foreach ($fields as $fieldName => $fieldValues) {
      if (isset($fieldValues['harmonize'])) {
        $displaySettings[$mode][$fieldName] = (bool) $fieldValues['harmonize'];
      }
    }

    // Set the new display settings and save.
    $settings->setDisplaySettings($displaySettings);
    $settings->save();
  }

}
