<?php

namespace Drupal\harmonize\Service;

use Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException;
use Drupal\Component\Plugin\Exception\PluginNotFoundException;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Datetime\DateFormatterInterface;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\image\Entity\ImageStyle;

/**
 * Service used in Refiners & Harmony files to access common helper functions.
 *
 * @package Drupal\harmonize\Service
 */
class Harmonics {

  /**
   * The Harmonize service injected through DI.
   *
   * @var Harmonize
   */
  protected Harmonize $harmonize;

  /**
   * The Cache service.
   *
   * @var CacheBackendInterface
   */
  protected CacheBackendInterface $cacheBackend;

  /**
   * The Date Formatter injected through DI.
   *
   * @var DateFormatterInterface
   */
  protected DateFormatterInterface $dateFormatter;

  /**
   * The Entity Type Manager injected through DI.
   *
   * @var EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * The Image Style storage.
   *
   * @var EntityStorageInterface
   */
  protected EntityStorageInterface $imageStyleStorage;

  /**
   * Service that contains all functionality for the harmonizer module.
   *
   * @param Harmonize $harmonize
   *   The groomer manager, injected through DI.
   * @param CacheBackendInterface $cacheBackend
   *   The Cache service injected through DI.
   * @param DateFormatterInterface $dateFormatter
   *   The Date Formatter object.
   * @param EntityTypeManagerInterface $entityTypeManager
   *   Entity Type manager injected through DI.
   */
  public function __construct(
    Harmonize                  $harmonize,
    CacheBackendInterface      $cacheBackend,
    DateFormatterInterface     $dateFormatter,
    EntityTypeManagerInterface $entityTypeManager
  ) {
    $this->harmonize = $harmonize;
    $this->cacheBackend = $cacheBackend;
    $this->dateFormatter = $dateFormatter;
    $this->entityTypeManager = $entityTypeManager;
    try {
      $this->imageStyleStorage = $this->entityTypeManager->getStorage(
        'image_style'
      );
    } catch (InvalidPluginDefinitionException | PluginNotFoundException $e) {
      $this->harmonize->loggerChannel->error("Unable to load Image Style Storage.");
    }
  }

  /**
   * Format an image into a standardized object appropriate for fractal templates.
   *
   * @param array $imageData
   *   The image data obtained from Harmonize.
   * @param string|null $style
   *   Image style to format the image with.
   *
   * @return array
   *   Formatted image data.
   */
  public function niceImage(array $imageData, string $style = NULL) : ?array {
    // Set the default image source.
    $src = $imageData['attributes']['src'];

    // Get image style if we must.
    if ($style) {
      /* @var ImageStyle $imageStyle */
      $imageStyle = $this->imageStyleStorage->load($style);

      if ($imageStyle === NULL) {
        $this->harmonize->loggerChannel->warning('The image style "' . $style . '" doesn\'t exist.');
      }
      else {
        $src = $imageStyle->buildUrl($imageData['uri']);
      }
    }

    return [
      'alt' => $imageData['attributes']['alt'],
      'src' => $src,
    ];
  }

  /**
   * Format a date into a standardized object appropriate for fractal templates.
   *
   * @param string|array $date
   *   The original date string.
   * @param string $format
   *   Specify the format the date label should be displayed in.
   *
   * @return array
   *   Date object containing label and readable values.
   */
  public function niceDate($date, string $format) : ?array {
    // Handle date ranges.
    if (is_array($date) && isset($date['from'], $date['to'])) {
      $from = $this->niceDate($date['from'], $format);
      $to = $this->niceDate($date['to'], $format);

      // If they're on the same day, we format our label differently.
      if ($from['readable'] === $to['readable']) {
        return [
          'label' => $from['label'] . ' | ' . $from['time'] . ' - ' . $to['time'] . ' ' . $from['timezone'],
          'readable' => $from['readable'],
        ];
      }

      return [
        'label' => $from['label'] . ' ' . $from['time'] . ' - ' . $to['label'] . ' ' . $to['time'] . ' ' . $from['timezone'],
        'readable' => $from['readable'],
      ];
    }

    return [
      'label'     => $this->dateFormatter->format(strtotime($date), 'custom', $format, date_default_timezone_get()),
      'readable'  => $this->dateFormatter->format(strtotime($date), 'custom', 'Y-m-d', date_default_timezone_get()),
      'time'      => $this->dateFormatter->format(strtotime($date), 'custom', 'h:i a', date_default_timezone_get()),
      'timezone'  => $this->dateFormatter->format(strtotime($date), 'custom', 'T', date_default_timezone_get()),
    ];
  }

  /**
   * Format a link into a standardized object appropriate for fractal templates.
   *
   * @param array $linkFieldConsignment
   *   Consigned data of the link field.
   * @param array $overrides
   *   Used to override attributes of the link field.
   *
   * @return array
   *   Array of normalized consistent data of the link.
   */
  function niceLink(array $linkFieldConsignment, array $overrides = []) : ?array {
    if (!$linkFieldConsignment) {
      return NULL;
    }

    $link = [
      'href'    => $linkFieldConsignment['attributes']['href'] ?? NULL,
      'text'    => $linkFieldConsignment['label'] ?? NULL,
      'target'  => $linkFieldConsignment['attributes']['target'] ?? '_self',
    ];

    if ($overrides) {
      $link = array_merge($link, $overrides);
    }

    if ($link['target'] === '_blank') {
      $link['rel'] = 'noopener noreferrer';
    }

    return $link;
  }
}
